"""Browser module."""

from __future__ import annotations
from saas.web.url import Url, InvalidUrlException
from html.parser import HTMLParser
from urllib.error import HTTPError
from saas.web.page import Page
import urllib.request


class Browser:
    """Browser class.

    The browser class retrieves web pages
    from urls.
    """

    @staticmethod
    def get_page(url: Url) -> Page:
        """Get page.

        Fetch page at url

        Args:
            url: Url page is located at

        Returns:
            The requested page if response is 200 otherwise None
            Page
        """
        parser = LinkParser()
        page = Page()
        try:
            with urllib.request.urlopen(url.to_string()) as response:
                html = response.read()
        except HTTPError as error:
            page.status_code = error.getcode()
            return page

        page.status_code = response.getcode()
        headers = dict(response.headers)
        for header in headers:
            if header.lower() == 'content-type':
                page.content_type = headers[header]

        links = parser.parse(str(html))

        for link in links:
            try:
                page.add_url(Url.from_string(link))
            except InvalidUrlException as e:
                try:
                    page.add_url(url.create_child_url(link))
                except InvalidUrlException as e:
                    pass
        return page


class LinkParser(HTMLParser):
    """Link parser."""

    def parse(self, html: str) -> list:
        """Parse links in html string.

        Args:
            html: html document

        Returns:
            List of the found links
            list
        """
        self.links = []  # type: list
        self.feed(html)
        return self.links

    def handle_starttag(self, tag, attrs):
        """Handle starttag.

        Called by parse method

        Args:
            tag: element tag
            attrs: element attributes
        """
        if tag == 'a':
            for name, value in attrs:
                if name == 'href':
                    self.links.append(value)



"""Console module.

Collection of console helpers.
"""

import sys
import beeprint


DEBUG = False


def pp(var):
    """Pretty print.

    Wrapper around beeprint

    Args:
        var: some variable
    """
    beeprint.pp(var)


def dd(var):
    """Pretty print and die.

    Args:
        var: some variable
    """
    pp(var)
    exit()


def p(var, end=None):
    """Print.

    Wrapper around print.

    Args:
        var: some variable
        end: end of line character (default: {None})
    """
    if end is not None:
        print(var, end=end)
    else:
        print(var)
    sys.stdout.flush()


def dcr(message: str):
    """Print crawler debug info.

    Args:
        message: debug message
    """
    if DEBUG:
        p('debug [crawler]: ' + message)


def dp(message: str):
    """Print photographer debug info.

    Args:
        message: debug message
    """
    if DEBUG:
        p('debug [photographer]: ' + message)


def dca(message: str):
    """Print camera debug info.

    Args:
        message: debug message
    """
    if DEBUG:
        p('debug [camera]: ' + message)


def df(message: str):
    """Print filesystem debug info.

    Args:
        message: debug message
    """
    if DEBUG:
        p('debug [filesystem]: ' + message)


def eol():
    """Print end of line character."""
    p('')



"""Files helper module."""

import errno
import os


def create_dir(directory: str) -> str:
    """Create directory.

    Args:
        directory: path to directory

    Returns:
        Absolute path to created directory
        str
    """
    path = real_path(directory)
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    return path


def real_path(path: str) -> str:
    """Real path.

    Args:
        path: relative or absolute path

    Returns:
        Absolute path to provided path
        str
    """
    if path[0] == '~':
        root = os.path.expanduser('~')
        path = path[1:]
    elif path[0] == '/':
        root = '/'
    else:
        root = os.getcwd()
    return f'{root}/{path}'.replace('//', '/').replace('//', '/')



"""Filesystem test."""

from saas.photographer.photo import PhotoPath, Screenshot
from saas.mount.file import Directory, File, LastCapture
from saas.storage.datadir import DataDirectory
from saas.mount.filesystem import Filesystem
from unittest.mock import MagicMock, call
import saas.storage.refresh as refresh
from saas.storage.index import Index
import saas.utils.console as console
from saas.web.url import Url
from os.path import dirname
import unittest
import time
import os


class TestFilesystem(unittest.TestCase):
    """Test filesystem class."""

    def setUp(self):
        """Set up test."""
        self.console = console
        self.refresh_rate = refresh.Hourly
        self.datadir = DataDirectory(dirname(__file__) + '/datadir')
        self.index = Index(self.datadir, MagicMock())
        self.filesystem = Filesystem(self.index, self.refresh_rate)

    def tearDown(self):
        """Tear down test."""
        self.datadir.remove_data_dir()

    def assertListOfFilesEqual(self, expected: list, actual: list):
        """Assert list of files equal.

        Args:
            expected: Expected list of files
            actual: Actual list of files
        """
        msg = 'Failed asserting list of files where equal expected'
        self.assertEqual(len(expected), len(actual), msg=msg)
        for i, file in enumerate(expected):
            self.assertEqual(file.filename, actual[i].filename, msg=msg)
            self.assertIsInstance(cls=file.__class__, obj=actual[i], msg=msg)

    def test_filesystem_can_list_contents_of_root_directory(self):
        """Test filesystem can list root directory."""
        self.index.photos_unique_domains = MagicMock(return_value=[
            'example.com',
            'example.net'
        ])

        files = self.filesystem._list('/')

        self.assertListOfFilesEqual(
            [
                Directory('.'),
                Directory('..'),
                Directory('example.com'),
                Directory('example.net'),
            ],
            files
        )
        self.index.photos_unique_domains.assert_called_with(self.refresh_rate)

    def test_filesystem_can_list_contents_of_domain(self):
        """Test filesystem can list contents of domain."""
        self.index.photos_unique_captures_of_domain = MagicMock(return_value=[
            '2019-01-13H20:00',
            '2019-01-13H21:00',
            '2019-01-13H22:00',
        ])

        expected = [
            Directory('.'),
            Directory('..'),
            Directory('2019-01-13H20:00'),
            Directory('2019-01-13H21:00'),
            Directory('2019-01-13H22:00'),
            Directory(LastCapture.FILENAME),
        ]

        files = self.filesystem._list('/example.com')
        self.assertListOfFilesEqual(expected, files)

        files = self.filesystem._list('/example.com/')
        self.assertListOfFilesEqual(expected, files)

        self.index.photos_unique_captures_of_domain.assert_called_with(
            'example.com',
            self.refresh_rate
        )

    def test_filesystem_can_list_contents_of_capture_at_given_path(self):
        """Test filesystem can list contents of capture at given path."""
        self.index.photos_list_files_in_directory = MagicMock(return_value=[
            'index.png',
            'contact.png',
            'about.png',
        ])
        self.index.photos_list_directories_in_directory = MagicMock(
            return_value=[
                'sub_dir_1',
                'sub_dir_2',
            ]
        )

        expected = [
            Directory('.'),
            Directory('..'),
            File('index.png'),
            File('contact.png'),
            File('about.png'),
            Directory('sub_dir_1'),
            Directory('sub_dir_2'),
        ]

        files = self.filesystem._list('/example.com/2019-01-13H20:00/')
        self.assertListOfFilesEqual(expected, files)

        files = self.filesystem._list('/example.com/2019-01-13H20:00')
        self.assertListOfFilesEqual(expected, files)

        files = self.filesystem._list('/example.com/2019-01-13H20:00/foo/bar/')
        self.assertListOfFilesEqual(expected, files)

        calls = [
            call('example.com', '2019-01-13H20:00', '/', self.refresh_rate),
            call('example.com', '2019-01-13H20:00', '/', self.refresh_rate),
            call(
                'example.com',
                '2019-01-13H20:00',
                '/foo/bar/',
                self.refresh_rate
            ),
        ]
        self.index.photos_list_files_in_directory.assert_has_calls(calls)
        self.index.photos_list_directories_in_directory.assert_has_calls(calls)

    def test_filesystem_can_get_attributes_of_directory(self):
        """Test filesystem can get attributes of directory."""
        time.time = MagicMock(return_value=time.time())
        self.index.photos_directory_exists = MagicMock(return_value=True)
        self.index.photos_unique_domains = MagicMock(
            return_value=['example.com']
        )
        self.index.photos_unique_captures_of_domain = MagicMock(
            return_value=['2019-01-13H20:00']
        )

        expected = {
            'st_atime': time.time(),
            'st_ctime': time.time(),
            'st_gid': os.getgid(),
            'st_mode': Directory('').ST_MODE,
            'st_mtime': time.time(),
            'st_size': 0,
            'st_uid': os.getuid(),
        }

        attr = self.filesystem._attributes('/')
        self.assertEqual(expected, attr)

        attr = self.filesystem._attributes('/example.com/')
        self.assertEqual(expected, attr)

        attr = self.filesystem._attributes('/example.com/2019-01-13H20:00')
        self.assertEqual(expected, attr)

        attr = self.filesystem._attributes('/example.com/2019-01-13H20:00/')
        self.assertEqual(expected, attr)

        attr = self.filesystem._attributes(
            '/example.com/2019-01-13H20:00/foo/bar'
        )
        self.assertEqual(expected, attr)
        self.index.photos_directory_exists.assert_called_with(
            domain='example.com',
            captured_at='2019-01-13H20:00',
            directory='/foo/bar/',
            refresh_rate=self.refresh_rate
        )

    def test_filesystem_can_get_attributes_of_file(self):
        """Test filesystem can get attributes of file."""
        time.time = MagicMock(return_value=time.time())
        self.index.photos_directory_exists = MagicMock(return_value=False)
        self.index.photos_file_exists = MagicMock(
            return_value=123000  # returns filesize
        )

        expected = {
            'st_atime': time.time(),
            'st_ctime': time.time(),
            'st_gid': os.getgid(),
            'st_mode': File('').ST_MODE,
            'st_mtime': time.time(),
            'st_size': 123000,
            'st_uid': os.getuid(),
        }

        attr = self.filesystem._attributes(
            '/example.com/2019-01-13H20:00/index.png'
        )
        self.assertEqual(expected, attr)
        self.index.photos_file_exists.assert_called_with(
            domain='example.com',
            captured_at='2019-01-13H20:00',
            full_filename='/index.png',
            refresh_rate=self.refresh_rate
        )

    def test_filesystem_can_translate_path_to_file_in_datadir(self):
        """Test filesystem can translate path to file in datadir."""
        datadir_path = PhotoPath(self.datadir)
        url = Url.from_string('https://example.com/foo/bar')
        photo = Screenshot(url, datadir_path, self.refresh_rate)
        self.index.es.index = MagicMock()
        photo.path.filesize = MagicMock(return_value=10000)
        self.index.save_photo(photo)

        self.index.photos_file_exists = MagicMock(return_value=123000)
        self.index.photos_get_photo = MagicMock(return_value=photo)

        path = self.filesystem._translate_path(
            '/example.com/2019-01-13H20:00/foo/bar.png'
        )
        self.assertEqual(datadir_path.full_path(), path)


if __name__ == '__main__':
    unittest.main()



"""Path test."""

from saas.mount.file import Path
import unittest


class TestPath(unittest.TestCase):
    """Test path class."""

    def test_path_can_parse_parts_of_path(self):
        """Test path can parse parts."""
        path = '/example.com/2019-01-13H20:00/foo/bar/baz.png'
        path = Path(path)
        self.assertEqual('example.com', path.domain)
        self.assertEqual('2019-01-13H20:00', path.captured_at)
        self.assertEqual('/foo/bar/baz.png', path.end)

    def test_path_can_parse_end_correctly(self):
        """Test path can parse end correctly."""
        path = Path('/example.com/2019-01-13H20:00')
        self.assertEqual('/', path.end)

        path = Path('/example.com/2019-01-13H20:00/')
        self.assertEqual('/', path.end)

        path = Path('/example.com/2019-01-13H20:00/foo')
        self.assertEqual('/foo', path.end)

        path = Path('/example.com/2019-01-13H20:00/foo/')
        self.assertEqual('/foo', path.end)

        path = Path('/example.com/2019-01-13H20:00/foo/bar')
        self.assertEqual('/foo/bar', path.end)

        path = Path('/example.com/2019-01-13H20:00/foo/bar/')
        self.assertEqual('/foo/bar', path.end)

        path = Path('/example.com/2019-01-13H20:00/foo/bar/baz.png')
        self.assertEqual('/foo/bar/baz.png', path.end)

    def test_check_for_domain(self):
        """Test check for domain."""
        path = Path('/example.com/2019-01-13H20:00')
        self.assertTrue(path.includes_domain())

        path = Path('/example.com/')
        self.assertTrue(path.includes_domain())

        path = Path('/example.com')
        self.assertTrue(path.includes_domain())

        path = Path('/')
        self.assertFalse(path.includes_domain())

    def test_check_for_captured_at(self):
        """Test check for captured_at."""
        path = Path('/example.com/2019-01-13H20:00/foo/bar')
        self.assertTrue(path.includes_captured_at())

        path = Path('/example.com/2019-01-13H20:00/foo')
        self.assertTrue(path.includes_captured_at())

        path = Path('/example.com/2019-01-13H20:00/')
        self.assertTrue(path.includes_captured_at())

        path = Path('/example.com/2019-01-13H20:00')
        self.assertTrue(path.includes_captured_at())

        path = Path('/example.com/')
        self.assertFalse(path.includes_captured_at())

        path = Path('/')
        self.assertFalse(path.includes_captured_at())

    def test_check_for_end(self):
        """Test check for end."""
        path = Path('/example.com/2019-01-13H20:00/foo/bar')
        self.assertTrue(path.includes_end())

        path = Path('/example.com/2019-01-13H20:00/foo')
        self.assertTrue(path.includes_end())

        path = Path('/example.com/2019-01-13H20:00/')
        self.assertFalse(path.includes_end())

        path = Path('/example.com/2019-01-13H20:00')
        self.assertFalse(path.includes_end())

        path = Path('/example.com/')
        self.assertFalse(path.includes_end())

        path = Path('/')
        self.assertFalse(path.includes_captured_at())

    def test_end_can_be_treated_as_file(self):
        """Test end can be treated as file."""
        path = Path('/example.com/2019-01-13H20:00/foo/bar')
        self.assertEqual('/foo/bar', path.end_as_file())

        path = Path('/example.com/2019-01-13H20:00/foo/bar/')
        self.assertEqual('/foo/bar', path.end_as_file())

        path = Path('/example.com/2019-01-13H20:00/')
        self.assertEqual('/', path.end_as_file())

    def test_end_can_be_treated_as_directory(self):
        """Test end can be treated as directory."""
        path = Path('/example.com/2019-01-13H20:00/foo/bar')
        self.assertEqual('/foo/bar/', path.end_as_directory())

        path = Path('/example.com/2019-01-13H20:00/foo/bar/')
        self.assertEqual('/foo/bar/', path.end_as_directory())

        path = Path('/example.com/2019-01-13H20:00/')
        self.assertEqual('/', path.end_as_directory())


if __name__ == '__main__':
    unittest.main()



"""Crawler module."""

from __future__ import annotations
from saas.storage.index import Index, EmptySearchResultException
import saas.utils.console as console
from saas.web.browser import Browser
from saas.web.url import Url
import time
import os


class Crawler:
    """Crawler.

    The crawler crawls the world wide web
    for websites.
    """

    def __init__(
        self,
        url_file: str,
        index: Index,
        ignore_found_urls: bool=False,
        stay_at_domain: bool=False,
    ):
        """Create crawler.

        Args:
            url_file: path to url file
            index: Index storage for queued urls
            ignore_found_urls: if crawler should ignore new urls found on
                pages it crawls
            stay_at_domain: if crawler should ignore urls from a different
                domain than the one it was found at
        """
        self.source_is_open = False
        self.source_path = ''
        self.source_mode = ''
        self._open_source('r', url_file)
        self.ignore_found_urls = ignore_found_urls
        self.stay_at_domain = stay_at_domain
        self.index = index

    def _open_source(self, mode: str, url_file: str=''):
        """Open source file.

        Args:
            mode: read or write
            url_file: path to url file

        Raises:
            UrlFileNotFoundError: if file wasn't found
        """
        if self.source_is_open and self.source_mode == mode:
            return
        if self.source_is_open and self.source_mode != mode:
            self._close_source()
        self.source_is_open = True

        if self.source_path == '':
            file = self._real_path(url_file)
            if not os.path.isfile(file):
                raise UrlFileNotFoundError(f'url file was not found at {file}')
            self.source_path = file

        self.source_mode = mode
        self.source = open(self.source_path, mode)

    def _close_source(self):
        """Close url source file."""
        self.source_is_open = False
        self.source.close()

    def _real_path(self, url_file: str) -> str:
        """Real path.

        Args:
            url_file: relative path

        Returns:
            Absolute oath
            str
        """
        if url_file[0] == '~':
            root = os.path.expanduser('~')
            url_file = url_file[1:]
        elif url_file[0] == '/':
            root = '/'
        else:
            root = os.getcwd()
        return f'{root}/{url_file}'.replace('//', '/').replace('//', '/')

    def tick(self):
        """Tick.

        Check if there are any uncrawled urls, if none exists
        then sleep for a second, otherwise crawl url.
        """
        url = self._next_url()

        if not url:
            time.sleep(1)
            return

        if url is not None:
            self.index.add_crawled_url(url)

            console.dcr(f'crawling {url.to_string()}')

            page = Browser.get_page(url)
            if 'text/html' not in page.content_type:
                page.status_code = 0

            console.dcr(f'{url.to_string()} responded with {page.status_code}')

            self.index.set_status_code_for_crawled_url(
                url,
                page.status_code
            )

            if self.ignore_found_urls:
                return

            if page.status_code != 200:
                return

            if self.stay_at_domain:
                page.remove_urls_not_from_domain(url.domain)

            console.dcr(f'found {len(page.urls)} links at {url.to_string()}')

            self.index.add_uncrawled_urls(page.urls)

    def _next_url(self):
        """Get next url to crawl.

        Returns:
            A url to crawl, None if no url was found
            Url or None
        """
        self._open_source('r')
        lines = sum(1 for line in self.source)

        if lines == 0:
            return self._next_url_in_index()

        self.source.seek(0)
        lines = self.source.read().split('\n')

        if lines[0] == '':
            return self._next_url_in_index()

        self._open_source('w')
        for line in lines:
            if line != lines[0] and line.strip() != '':
                self.source.write(line + '\n')
        self._close_source()

        url = Url.from_string(lines[0])
        return url

    def _next_url_in_index(self):
        """Get next url to crawl from index.

        Returns:
            A url to crawl, None if no url was found
            Url or None
        """
        try:
            url = self.index.random_uncrawled_url()
        except EmptySearchResultException:
            return None
        self.index.remove_uncrawled_url(url.hash())
        return url

    def stop(self):
        """Stop crawler."""
        self._close_source()


class UrlFileNotFoundError(ValueError):
    """Url file not found error."""

    pass



"""Index test."""

from saas.photographer.photo import PhotoPath, LoadingPhoto
from saas.storage.datadir import DataDirectory
from saas.photographer.photo import Photo
import saas.storage.refresh as refresh
from saas.storage.index import Index
from unittest.mock import MagicMock
from saas.web.url import Url
from os.path import dirname
import unittest
import time


class TestIndex(unittest.TestCase):
    """Test index class."""

    def setUp(self):
        """Set up test."""
        self.datadir = DataDirectory(dirname(__file__) + '/datadir')
        self.index = Index(self.datadir, MagicMock())

    def tearDown(self):
        """Tear down test."""
        self.datadir.remove_data_dir()

    def search_returns_doc(self, doc: dict):
        """Search to elastic search returns doc.

        Mock search method of self.index.es to return given doc

        Args:
            doc: document or partial document to return
        """
        self.index.es.search = MagicMock(return_value={
            'hits': {
                'total': 1,
                'hits': [
                    doc
                ]
            }
        })

    def search_returns_aggregation(self, index: str, buckets: list):
        """Search to elastic search returns aggregation.

        Args:
            index: index being searched
            buckets: buckets that's returned
        """
        self.index.es.search = MagicMock(return_value={
            'aggregations': {
                index: {
                    'buckets': buckets
                }
            }
        })

    def test_recently_crawled_url_can_be_fetched(self):
        """Test recently crawled url can be fetched."""
        self.search_returns_doc({
            '_id': 'xxx...',
            '_source': {
                'url': 'http://example.com',
                'timestamp': 1547229873.257901
            }
        })

        url = self.index.recently_crawled_url(refresh.Hourly)
        self.assertIsInstance(cls=Url, obj=url)
        self.assertEqual('http://example.com', url.to_string())
        self.index.es.search.assert_called_with(
            index='crawled',
            size=5,
            body={
                'query': {
                    'bool': {
                        'must': {
                            'term': {
                                'status_code': 200,
                            }
                        },
                        'must_not': [
                            {
                                'term': {
                                    'lock_value': refresh.Hourly().lock(),
                                }
                            }
                        ]
                    }
                },
                'sort': [
                    {
                        'timestamp': {
                            'order': 'desc'
                        }
                    }
                ]
            }
        )

    def test_lock_can_be_placed_on_crawled_url(self):
        """Test lock can be placed on crawled url."""
        url = Url.from_string('http://example.com')
        self.index.es.update = MagicMock()

        self.index.lock_crawled_url(url, refresh.Hourly)
        self.index.es.update.assert_called_with(
            index='crawled',
            doc_type='url',
            id=url.hash(),
            retry_on_conflict=3,
            body={
                'doc': {
                    'lock_format': refresh.Hourly.lock_format(),
                    'lock_value': refresh.Hourly().lock(),
                }
            }
        )

    def test_index_can_store_photo(self):
        """Test index can store a photo."""
        self.index.es.index = MagicMock()
        time.time = MagicMock(return_value=time.time())

        url = Url.from_string('http://example.com')
        path = PhotoPath(self.datadir)
        path.filesize = MagicMock(return_value=10000)

        photo = LoadingPhoto(
            url=url,
            path=path,
            refresh_rate=refresh.Hourly
        )

        self.index.save_photo(photo)
        self.index.es.index.assert_called_with(
            index='photos',
            doc_type='photo',
            id=path.uuid,
            body={
                'url_id': url.hash(),
                'refresh_rate': refresh.Hourly.lock_format(),
                'captured_at': refresh.Hourly().lock(),
                'filesize': photo.filesize(),
                'filename': photo.filename(),
                'directory': photo.directory(),
                'domain': photo.domain(),
                'timestamp': int(time.time())
            }
        )

    def test_index_can_list_unique_photo_domains(self):
        """Test index can list unique photos."""
        self.search_returns_aggregation('photos', [
            {
                'key': 'example.com',
            },
            {
                'key': 'example.net',
            }
        ])

        domains = self.index.photos_unique_domains(refresh.Hourly)

        self.assertEqual(['example.com', 'example.net'], domains)
        self.index.es.search.assert_called_with(
            index='photos',
            size=0,
            body={
                'query': {
                    'bool': {
                        'must': {
                            'term': {
                                'refresh_rate': refresh.Hourly.lock_format(),
                            }
                        },
                    }
                },
                'aggs': {
                    'photos': {
                        'terms': {
                            'field': 'domain',
                            'size': 10000
                        }
                    }
                }
            }
        )

    def test_index_can_list_unique_captures_of_domains(self):
        """Test index can list unique captures of domain."""
        self.search_returns_aggregation('photos', [
            {
                'key': '2019-01-13H20:00',
            },
            {
                'key': '2019-01-13H21:00',
            }
        ])

        domains = self.index.photos_unique_captures_of_domain(
            'example.com',
            refresh.Hourly
        )

        format = refresh.Hourly.lock_format()
        self.assertEqual(['2019-01-13H20:00', '2019-01-13H21:00'], domains)
        self.index.es.search.assert_called_with(
            index='photos',
            size=0,
            body={
                'query': {
                    'bool': {
                        'must': [
                            {
                                'term': {
                                    'domain': 'example.com',
                                }
                            },
                            {
                                'term': {
                                    'refresh_rate': format,
                                }
                            }
                        ],
                    }
                },
                'aggs': {
                    'photos': {
                        'terms': {
                            'field': 'captured_at',
                            'size': 10000
                        }
                    }
                }
            }
        )

    def test_photo_can_be_retrieved(self):
        """Test photo can be retrieved."""
        format = refresh.Hourly.lock_format()
        capture = refresh.Hourly().lock()
        self.search_returns_doc({
            '_id': 'uuid-xxx...',
            '_source': {
                'url_id': 'xxx...',
                'refresh_rate': format,
                'captured_at': capture,
                'filename': 'some-filename.png',
                'directory': '/some/path/',
                'domain': 'example.com',
                'filesize': 12300,
                'timestamp': time.time(),
            }
        })

        photo = self.index.photos_get_photo(
            domain='example.com',
            captured_at=capture,
            full_filename='/some/path/some-filename.png',
            refresh_rate=refresh.Hourly
        )

        self.assertIsInstance(cls=Photo, obj=photo)
        self.assertEqual('uuid-xxx...', photo.path.uuid)
        self.index.es.search.assert_called_with(
            index='photos',
            size=1,
            body={
                'query': {
                    'bool': {
                        'must': [
                            {
                                'term': {
                                    'domain': 'example.com',
                                }
                            },
                            {
                                'term': {
                                    'refresh_rate': format,
                                }
                            },
                            {
                                'term': {
                                    'captured_at': capture,
                                }
                            },
                            {
                                'term': {
                                    'directory': '/some/path/',
                                }
                            },
                            {
                                'term': {
                                    'filename': 'some-filename.png',
                                }
                            }
                        ],
                    }
                }
            }
        )

    def test_directories_within_a_directory_can_be_fetched(self):
        """Test directories within a directory can be fetched."""
        format = refresh.Hourly.lock_format()
        capture = refresh.Hourly().lock()
        self.search_returns_aggregation('photos', [
            {
                'key': '/path/to/some/dir/',
            },
            {
                'key': '/path/to/some/other/dir/',
            },
            {
                'key': '/path/to/not/same/other/dir/',
            },
            {
                'key': '/path/to/a/dir/',
            }
        ])

        directories = self.index.photos_list_directories_in_directory(
            domain='example.com',
            captured_at=capture,
            directory='/path/to/',
            refresh_rate=refresh.Hourly
        )

        self.assertEqual(['some', 'not', 'a'], directories)
        self.index.es.search.assert_called_with(
            index='photos',
            size=0,
            body={
                'query': {
                    'bool': {
                        'must': [
                            {
                                'term': {
                                    'domain': 'example.com',
                                }
                            },
                            {
                                'term': {
                                    'refresh_rate': format,
                                }
                            },
                            {
                                'term': {
                                    'captured_at': capture,
                                }
                            },
                            {
                                'wildcard': {
                                    'directory': '/path/to/*',
                                }
                            }
                        ],
                    }
                },
                'aggs': {
                    'photos': {
                        'terms': {
                            'field': 'directory',
                            'size': 10000
                        }
                    }
                }
            }
        )


if __name__ == '__main__':
    unittest.main()



"""Javascript module."""

from __future__ import annotations
from os.path import dirname


class JavascriptSnippets:
    """Javascript snippets class."""

    DOCUMENT_HEIGHT = ''

    IMAGES_MONITOR = ''

    IMAGES_LOADED = ''

    SCROLL_Y_AXIS = ''

    IMAGE_COUNT = ''

    SCRIPT_COUNT = ''

    STYLESHEET_COUNT = ''

    @staticmethod
    def load():
        """Load javscript snippets."""
        JavascriptSnippets.DOCUMENT_HEIGHT = JavascriptSnippets._load_snippet(
            'document_height.js'
        )
        JavascriptSnippets.IMAGES_MONITOR = JavascriptSnippets._load_snippet(
            'images_monitor.js'
        )
        JavascriptSnippets.IMAGES_LOADED = JavascriptSnippets._load_snippet(
            'images_loaded.js'
        )
        JavascriptSnippets.SCROLL_Y_AXIS = JavascriptSnippets._load_snippet(
            'scroll_y.js'
        )
        JavascriptSnippets.IMAGE_COUNT = JavascriptSnippets._load_snippet(
            'image_count.js'
        )
        JavascriptSnippets.SCRIPT_COUNT = JavascriptSnippets._load_snippet(
            'script_count.js'
        )
        JavascriptSnippets.STYLESHEET_COUNT = JavascriptSnippets._load_snippet(
            'stylesheet_count.js'
        )

    def _load_snippet(filename) -> str:
        """Load snippet from file.

        Args:
            filename: name of script eg. filename.js

        Returns:
            The content of the script file
            str
        """
        fullpath = f'{dirname(__file__)}/js/{filename}'
        file = open(fullpath, 'r')
        script = file.read()
        file.close()
        return script



"""Arguments module."""

from pkg_resources import get_distribution
import argparse


def get_argument_parser():
    """Get argument parser."""
    parser = argparse.ArgumentParser(
        prog='saas',
        description='Screenshot as a service',
        formatter_class=argparse.HelpFormatter,
    )

    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {}'.format(get_distribution('saas').version),
    )

    parser.add_argument(
        '--debug',
        action='store_true',
        default=False,
        help='Display debugging information',
    )

    parser.add_argument(
        'url_file',
        type=str,
        default=None,
        help='Path to input url file',
    )

    parser.add_argument(
        'mountpoint',
        type=str,
        default=None,
        help='Where to mount filesystem via FUSE',
    )

    parser.add_argument(
        '--refresh-rate',
        metavar='',
        type=str,
        default='hour',
        choices=['day', 'hour', 'minute'],
        help='''
            Refresh captures of urls every 'day', 'hour'
            or 'minute' (default: %(default)s)
        ''',
    )

    parser.add_argument(
        '--crawler-threads',
        metavar='',
        type=int,
        default=1,
        help='''
            Number of crawler threads, usually not
            neccessary with more than one (default: %(default)s)
        ''',
    )

    parser.add_argument(
        '--photographer-threads',
        metavar='',
        type=int,
        default=1,
        help='''
            Number of photographer threads, beaware that
            increasing too much won't neccessarily speed up
            performance and hog the system (default: %(default)s)
        '''
    )

    parser.add_argument(
        '--data-dir',
        metavar='',
        type=str,
        default='~/.saas-data-dir',
        help='Path to data directory (default: %(default)s)'
    )

    parser.add_argument(
        '--clear-data-dir',
        action='store_true',
        default=False,
        help='Use flag to clear data directory on start',
    )

    parser.add_argument(
        '--elasticsearch-host',
        metavar='',
        type=str,
        default='localhost:9200',
        help='Elasticsearch host (default: %(default)s)',
    )

    parser.add_argument(
        '--setup-elasticsearch',
        action='store_true',
        default=False,
        help='Use flag to create indices in elasticsearch',
    )

    parser.add_argument(
        '--clear-elasticsearch',
        action='store_true',
        default=False,
        help='''
            Use flag to clear elasticsearch on start,
            WARNING: this will clear all indices found
            in elasticsearch instance
        ''',
    )

    parser.add_argument(
        '--stay-at-domain',
        action='store_true',
        default=False,
        help='''
            Use flag to ignore urls from a different
            domain than the one it was found at
        ''',
    )

    parser.add_argument(
        '--ignore-found-urls',
        action='store_true',
        default=False,
        help='Use flag to ignore urls found on crawled urls',
    )

    parser.add_argument(
        '--viewport-width',
        metavar='',
        type=int,
        default=1920,
        help='Width of camera viewport in pixels (default: %(default)s)',
    )

    parser.add_argument(
        '--viewport-height',
        metavar='',
        type=int,
        default=0,
        help='''
            Height of camera viewport in pixels, if set to 0 camera will
            try to take a full height high quality screenshot,
            which is way slower than fixed size (default: %(default)s)
        ''',
    )

    parser.add_argument(
        '--viewport-max-height',
        metavar='',
        type=int,
        help='''
            Max height of camera viewport in pixels, if --viewport-height
            is set this will be ignored
        ''',
    )

    parser.add_argument(
        '--optimize-storage',
        action='store_true',
        default=False,
        help='''
            Image files should be optimized to take up
            less storage (takes longer time to render)
        ''',
    )

    parser.add_argument(
        '--stop-if-idle',
        metavar='',
        type=int,
        default=0,
        help='''
            If greater than 0 saas will stop if it is idle for more
            than the provided number of minutes
        ''',
    )

    return parser



"""Photographer module."""

from __future__ import annotations
from saas.photographer.photo import PhotoPath, LoadingPhoto
from saas.storage.index import EmptySearchResultException
from saas.storage.datadir import DataDirectory
import saas.storage.refresh as refresh
import saas.photographer.camera as c
import saas.utils.console as console
from saas.storage.index import Index
from typing import Type, Optional
import saas.threads as threads
from saas.web.url import Url
import random
import time


class Photographer:
    """Photographer class."""

    def __init__(
        self,
        index: Index,
        refresh_rate: Type[refresh.RefreshRate],
        datadir: DataDirectory,
        viewport_width: int=1920,
        viewport_height: int=0,
        viewport_max_height: Optional[int]=None
    ):
        """Create new photographer.

        Args:
            index: Index where crawled urls are stored and
                photos should be indexed.
            refresh_rate: How often photographs should be refreshed,
                more exactly defines which lock should be placed on
                crawled urls
            datadir: Data directory to store pictures in
            viewport_width: width of camera viewport
            viewport_height: height of camera viewport
            viewport_max_height: max height of camera viewport
        """
        self.index = index
        self.refresh_rate = refresh_rate
        self.datadir = datadir
        self.viewport_width = viewport_width
        self.viewport_height = viewport_height
        self.viewport_max_height = viewport_max_height

    def tick(self):
        """Tick.

        Checkout a url from index, take photo of url,
        save to datadir and update index with photo
        metadata
        """
        try:
            timer = time.time()
            url = self._checkout_url()

            console.dp(f'taking photo of {url.to_string()}')

            path = PhotoPath(self.datadir)
            photo = LoadingPhoto(
                url=url,
                path=path,
                refresh_rate=self.refresh_rate
            )
            photo.save_loading_text()
            self.index.save_photo(photo)

            camera = c.Camera(
                viewport_width=self.viewport_width,
                viewport_height=self.viewport_height,
                viewport_max_height=self.viewport_max_height,
            )
            photo = camera.take_picture(url, path, self.refresh_rate)
            self.index.save_photo(photo)

            timer = int(time.time() - timer)
            console.p(
                f'photo was taken of {url.to_string()} took: {timer}s'
            )

        except EmptySearchResultException as e:
            pass
        finally:
            time.sleep(1)

    def _checkout_url(self) -> Url:
        """Checkout url.

        A checkout pulls the most recently added url from
        the "crawled" index. A lock is placed on the url
        for the given refresh rate.

        Returns:
            A url ready to take a picture of
            Url
        """
        crawled_urls = self.index.crawled_urls_count(self.refresh_rate)
        processes = threads.Controller.PHOTOGRAPHER_PROCESSES
        if crawled_urls < processes and processes > 1:
            # to prevent checkout of same url when number of urls
            # to take photographs of are small
            time.sleep(round(random.uniform(10, 0), 2))

        url = self.index.recently_crawled_url(self.refresh_rate)  # type: Url
        self.index.lock_crawled_url(url, self.refresh_rate)
        return url



"""Refresh rate module."""

from __future__ import annotations
from abc import ABCMeta, abstractmethod
import datetime


class RefreshRate(metaclass=ABCMeta):
    """Refresh class."""

    @staticmethod
    @abstractmethod
    def lock_format():
        """Get the human readable format of lock."""
        pass

    @abstractmethod
    def _lock_datetime_format(self):
        """Get the format of lock used to make lock."""
        pass

    def lock(self) -> str:
        """Make lock.

        Returns:
            Lock to store in index.
            str
        """
        return datetime.datetime.today().strftime(self._lock_datetime_format())


class Daily(RefreshRate):
    """Daily refresh.

    Refresh photographs every day.
    """

    @staticmethod
    def lock_format():
        """Get the human readable format of lock."""
        return 'daily'

    def _lock_datetime_format(self):
        """Get the format of lock used to make lock."""
        return '%Y%m%d'


class Hourly(RefreshRate):
    """Hourly refresh.

    Refresh photographs every hour.
    """

    @staticmethod
    def lock_format():
        """Get the human readable format of lock."""
        return 'hourly'

    def _lock_datetime_format(self):
        """Get the format of lock used to make lock."""
        return '%Y%m%d%H'


class EveryMinute(RefreshRate):
    """Refresh every minute.

    Refresh photographs every minute.
    """

    @staticmethod
    def lock_format():
        """Get the human readable format of lock."""
        return 'minute'

    def _lock_datetime_format(self):
        """Get the format of lock used to make lock."""
        return '%Y%m%d%H%M'



"""Data directory module."""

from __future__ import annotations
import saas.photographer.photo as PhotoPath
from saas.utils.files import create_dir
import saas.utils.console as console
import shutil
import os


class DataDirectory:
    """Data directory class."""

    def __init__(self, path: str, optimize_storage: bool=False):
        """Create new data directory instance.

        Args:
            path: path to root of data directory
            optimize_storage: if photo files should be optimized
        """
        self.root = path
        self.root = create_dir(self.root)
        self.optimize_storage = optimize_storage

    def remove_data_dir(self):
        """Remove data directory root."""
        shutil.rmtree(self.root)

    def clear(self):
        """Clear data directory."""
        console.p(f'clearing data directory at: {self.root}')
        shutil.rmtree(self.root)
        create_dir(self.root)

    def path_for_photo(self, photo_path: PhotoPath.PhotoPath) -> str:
        """Get path for photo in data directory.

        Args:
            photo_path: A photo path object

        Returns:
            An absolute path to the file in data dir where photo is stored
            str
        """
        first_char = photo_path.uuid[0]
        second_char = photo_path.uuid[1]
        directory = self.root + '/' + first_char + second_char + '/'
        if not os.path.exists(directory):
            os.mkdir(directory)
        return directory + photo_path.uuid + '.png'

    def optimize_file(self, path: str):
        """Optimize file at path.

        Args:
            path: full path to file in data directory

        Raises:
            MissingDependencyException: if imagemagick is not installed
        """
        if os.system('command -v mogrify > /dev/null') != 0:
            raise MissingDependencyException('missing dependency imagemagick')

        cmd = 'mogrify -filter Triangle -define filter:support=2'
        cmd += ' -unsharp 0.25x0.08+8.3+0.045 -dither None -posterize 136'
        cmd += ' -quality 82 -define png:compression-filter=5'
        cmd += ' -define png:compression-level=9'
        cmd += ' -define png:compression-strategy=1'
        cmd += ' -define png:exclude-chunk=all -interlace none'
        cmd += ' -colorspace sRGB'
        cmd += ' ' + path

        os.system(cmd)


class MissingDependencyException(Exception):
    """Missing dependency exception."""

    pass



"""Url test."""

from saas.web.url import Url
import unittest
import hashlib


class TestUrl(unittest.TestCase):
    """Test url class."""

    def test_url_can_be_created_from_string(self):
        """Test url can be created from string."""
        url = 'https://example.com/foo/bar/baz?bax=bat&moo=loo#hello'
        url = Url.from_string(url)
        self.assertEqual('https', url.scheme)
        self.assertEqual('example.com', url.domain)
        self.assertEqual('/foo/bar/baz', url.path)
        self.assertEqual('bax=bat&moo=loo', url.query)
        self.assertEqual('hello', url.fragment)

    def test_url_can_be_converted_back_to_string(self):
        """Test url can be converted back to string."""
        url = 'https://example.com/foo/bar/baz?bax=bat&moo=loo#hello'
        parsed = Url.from_string(url)
        self.assertEqual(url, parsed.to_string())

    def test_url_can_be_hashed(self):
        """Test url can be hashed."""
        url = 'https://example.com/foo/bar/baz?bax=bat&moo=loo#hello'
        hash = hashlib.sha256(url.encode()).hexdigest()

        parsed = Url.from_string(url)
        self.assertEqual(hash, parsed.hash())

    def test_fully_qualified_url_can_be_created_from_uri(self):
        """Test fully qualified url can be created from a uri and url."""
        root = 'https://example.com/foo'
        root = Url.from_string(root)

        self.assertEqual(
            'https://example.com/foo/bar',
            root.create_child_url('bar').to_string()
        )

        self.assertEqual(
            'https://example.com/bar',
            root.create_child_url('/bar').to_string()
        )

        self.assertEqual(
            'https://foo.com',
            root.create_child_url('//foo.com').to_string()
        )

        root = 'https://example.com?foo=bar'
        root = Url.from_string(root)

        self.assertEqual(
            'https://example.com/bar#foo',
            root.create_child_url('/bar#foo').to_string()
        )

    def test_trailing_slash_is_removed_from_path(self):
        """Test trailing slash of path is treated the same as without slash."""
        url1 = 'https://example.com/foo/bar/'  # with trailing slash
        url2 = 'https://example.com/foo/bar'  # without trailing slash
        url1 = Url.from_string(url1)
        url2 = Url.from_string(url2)

        self.assertEqual('https://example.com/foo/bar', url1.to_string())
        self.assertEqual('https://example.com/foo/bar', url2.to_string())

    def test_filename_can_be_created(self):
        """Test filename can be created from url."""
        url = Url.from_string('https://example.com/foo/BAR-baz-bAt')
        self.assertEqual('bar-baz-bat.png', url.make_filename())

        url = Url.from_string('https://example.com/')
        self.assertEqual('index.png', url.make_filename())

        url = Url.from_string('https://example.com?someid=123&otherid=FOO')
        self.assertEqual('someid-123-and-otherid-foo.png', url.make_filename())

        url = Url.from_string('https://example.com?someid=123456#foobarbaz')
        self.assertEqual('foobarbaz.png', url.make_filename())

    def test_directory_can_be_created(self):
        """Test directory can be created from url."""
        url = Url.from_string('https://example.com/foo/baz/BAR-123/index.html')
        self.assertEqual('/foo/baz/bar-123/', url.make_directory())

        url = Url.from_string('https://example.com')
        self.assertEqual('/', url.make_directory())

        url = Url.from_string('https://example.com/')
        self.assertEqual('/', url.make_directory())

        url = Url.from_string('https://example.com/foo?someid=123&otherid=FOO')
        self.assertEqual('/foo/?/', url.make_directory())

        url = Url.from_string('https://example.com/1/2?id=123&param#foo')
        self.assertEqual('/1/2/?/id-123-and-param/#/', url.make_directory())


if __name__ == '__main__':
    unittest.main()



"""Camera module."""

from __future__ import annotations
from urllib3.exceptions import ProtocolError, MaxRetryError
from saas.photographer.javascript import JavascriptSnippets
from selenium.common.exceptions import JavascriptException
from saas.photographer.photo import PhotoPath, Screenshot
from saas.storage.refresh import RefreshRate
from http.client import RemoteDisconnected
import saas.utils.console as console
from typing import Type, Optional
from selenium import webdriver
import saas.threads as threads
from saas.web.url import Url
import pkg_resources
import time
import os


class Camera:
    """Camera class."""

    def __init__(
        self,
        viewport_width: int=1920,
        viewport_height: int=0,
        viewport_max_height: Optional[int]=None
    ):
        """Create new camera.

        Args:
            viewport_width: width of camera viewport (default: {1920})
            viewport_height: height of camera viewport, if set to 0 camera
                will try to take a full height high quality screenshot,
                which is way slower than fixed size (default: {0})
            viewport_max_height: max height of camera viewport if
                viewport_height is not default value this will be ignored
        """
        self.webdriver = None  # type: webdriver.FirefoxProfile
        self.width = 0
        self.height = 0
        self.viewport_width = viewport_width
        self.viewport_height = viewport_height
        self.viewport_max_height = viewport_max_height

    def take_picture(
        self, url: Url,
        path: PhotoPath,
        refresh_rate: Type[RefreshRate]
    ) -> Screenshot:
        """Take picture of url.

        Uses the selenium webdriver to load url in firefox,
        make sure the entire page and it's assets are loaded
        then write it to data directory as a png.

        Args:
            url: Url to take picture of
            path: Path to store url at
            refresh_rate: Refresh rate for photo

        Returns:
            A picture of the given url
            Screenshot
        """
        os.environ['MOZ_HEADLESS'] = '1'
        try:
            console.dca('launching firefox, camera: {}x{}'.format(
                self.viewport_width,
                self.viewport_height if self.viewport_height != 0 else 'full'
            ))

            profile = self._create_webdriver_profile()
            self.webdriver = self._create_webdriver(
                profile,
                UserAgents.GOOGLEBOT
            )
            self._install_webdriver_addons()

            threads.Controller.webdrivers.append(
                self.webdriver.service.process.pid
            )

            console.dca(f'routing camera to {url.to_string()}')

            self._route_to_blank()
            self._route(url)

            if self.viewport_height != 0:
                # fixed height
                console.dca('taking fixed screenshot')
                self._set_resolution(self.viewport_width, self.viewport_height)
                self._start_images_monitor()
                self._wait_for_images_to_load()
            else:
                # fullpage screenshot
                console.dca('taking fullpage screenshot')
                self._set_resolution(self.viewport_width, 1080)
                self._start_images_monitor()
                self._wait_for_images_to_load()

                # scroll down the page to trigger load of images
                console.dca('making sure all images have loaded')
                steps = int(self._document_height() / 800)
                for i in range(1, steps):
                    scroll_to = i * 800
                    if self.viewport_max_height is not None:
                        if scroll_to >= self.viewport_max_height:
                            break
                    self._scroll_y_axis(scroll_to)
                    self._wait_for_images_to_load()

                # resize the viewport and make sure that it's scrolled
                # all the way to the top
                console.dca(f'resizing camera viewport for {url.to_string()}')
                self._scroll_y_axis(self._document_height() * -1)
                if self.viewport_max_height is None:
                    height = self._document_height()
                elif self._document_height() > self.viewport_max_height:
                    height = self.viewport_max_height
                else:
                    height = self._document_height()
                self._set_resolution(self.viewport_width, height)
                self._wait_for_images_to_load()
                self._scroll_y_axis(-500)
                self._wait_for_resize()

            console.dca(f'saving screenshot of {url.to_string()}')
            self._save(path)
        except RemoteDisconnected:
            pass
        except ProtocolError:
            pass
        except MaxRetryError:
            pass
        finally:
            if self.webdriver:
                self.webdriver.quit()

        if path.should_optimize():
            console.dca(f'optimizing screenshot of {url.to_string()}')
            timer = time.time()
            path.optimize()
            seconds = int(time.time() - timer)
            console.dca(f'optimizing of {url.to_string()} took {seconds}s')

        return Screenshot(
            url=url,
            path=path,
            refresh_rate=refresh_rate
        )

    def _create_webdriver_profile(self) -> webdriver.FirefoxProfile:
        """Create webdriver profile.

        Returns:
            An empty profile Firefox Profile to use when
            creating webdriver.
            webdriver.FirefoxProfile
        """
        return webdriver.FirefoxProfile()

    def _create_webdriver(
        self,
        profile: webdriver.FirefoxProfile,
        user_agent: str
    ) -> webdriver.Firefox:
        """Create webdriver.

        Args:
            profile: FirefoxProfile with configuration options
            user_agent: User agent string to use

        Returns:
            A webdriver that can be used to interact with the firefox browser
            webdriver.Firefox
        """
        profile.set_preference('general.useragent.override', user_agent)
        profile.set_preference('dom.popup_maximum', 0)
        profile.set_preference('privacy.popups.showBrowserMessage', False)
        profile.set_preference('dom.push.enabled', False)
        return webdriver.Firefox()

    def _install_webdriver_addons(self):
        """Install webdriver addons.

        The extensions that are installed either improve rendering
        performance. Or let firefox bypass paywalls or cookie/GDPR
        concent forms.
        """
        self.webdriver.install_addon(Addons.REFERER_HEADER, temporary=True)
        self.webdriver.install_addon(Addons.IDCAC)
        self.webdriver.install_addon(Addons.UBLOCK_ORIGIN, temporary=True)

    def _route(self, url: Url):
        """Route camera to url.

        Args:
            url: A Url to route camera to
        """
        self.webdriver.get(url.to_string())

    def _route_to_blank(self):
        """Route to blank page."""
        self.webdriver.get('about:blank')

    def _save(self, path: PhotoPath):
        """Save screen shot.

        Args:
            path: PhotoPath object used to retrieve path in data directory
            to save png file in
        """
        self.webdriver.save_screenshot(path.full_path())

    def _set_resolution(self, width: int, height: int):
        """Set camera resolution.

        Args:
            width: Width in pixels
            height: Height in pixels
        """
        self.width = width
        self.height = height
        self.webdriver.set_window_size(width, height)

    def _execute_script(self, script: str, retry: int=40):
        """Execute script in browser.

        Args:
            script: javascript to execute
            retry: number of times script can fail and retry

        Returns:
            Result of script
        """
        try:
            return self.webdriver.execute_script(script)
        except JavascriptException as e:
            if retry < 1:
                raise e
            time.sleep(0.75)
            return self._execute_script(script, retry - 1)

    def _document_height(self) -> int:
        """Get document height from webdriver.

        Returns:
            The height of the document at the page routed to
            int
        """
        height = self._execute_script(
            JavascriptSnippets.DOCUMENT_HEIGHT
        )  # type: int
        return height

    def _image_count(self) -> int:
        """Get number of images on page.

        Returns:
            The number of images the webpage has requested
            int
        """
        images = self._execute_script(
            JavascriptSnippets.IMAGE_COUNT
        )  # type: int
        return images

    def _script_count(self) -> int:
        """Get number of scripts page uses.

        Returns:
            The number of scripts the page has requested
            int
        """
        scripts = self._execute_script(
            JavascriptSnippets.SCRIPT_COUNT
        )  # type: int
        return scripts

    def _stylesheet_count(self) -> int:
        """Get number of stylesheets page uses.

        Returns:
            The number of stylesheets the page has requested
            int
        """
        stylesheets = self._execute_script(
            JavascriptSnippets.STYLESHEET_COUNT
        )  # type: int
        return stylesheets

    def _scroll_y_axis(self, pixels: int):
        """Scroll page on the y axis.

        Args:
            pixels: The number of pixels to scroll vertically
        """
        self._execute_script(
            JavascriptSnippets.SCROLL_Y_AXIS.replace('Y_PIXELS', str(pixels))
        )

    def _start_images_monitor(self):
        """Start images monitor script on page."""
        self._execute_script(
            JavascriptSnippets.IMAGES_MONITOR
        )

    def _wait_for_images_to_load(self):
        """Wait for all images to load on page."""
        while self._execute_script(
            JavascriptSnippets.IMAGES_LOADED
        ) is None:
            time.sleep(Limits.SLEEP_BETWEEN_IMAGE_LOAD_CHECK)

        try:
            while self._execute_script(
                JavascriptSnippets.IMAGES_LOADED
            ) < Limits.ALL_IMAGES_LOADED:
                time.sleep(Limits.SLEEP_BETWEEN_IMAGE_LOAD_CHECK)
        except TypeError:
            self._start_images_monitor()

    def _wait_for_resize(self):
        """Wait for resoultion resize to be completed.

        When the camera resizes and the entire page is shown
        in the viewport all assets will be loaded. This takes
        longer time to complete for more complex pages.

        Some examples:

            Tabloids (generally filled with crap)
            https://www.dailymail.co.uk/        score: 321
            https://www.expressen.se/           score: 124
            https://aftonbladet.se/             score: 35

            Image heavy news sites
            https://dn.se/                      score: 39
            https://www.di.se/                  score: 44
            https://svd.se/                     score: 31

            Non image heavy news sites
            https://wsj.com/                    score: 12
            https://www.nytimes.com/            score: 10

            Blogs and non image + script heavy news sites
            https://www.ft.com/                 score: 3
            https://daringfireball.net/         score: 3
            https://www.recode.net/             score: 7
            https://www.schneier.com/           score: 2

            Zero or next to zero images
            https://news.ycombinator.com/       score: 0
            https://example.com/                score: 0

            * These are only examples and may change,
              captured at Sun Jan 13 00:37:26 CET 2019

        """
        images = self._image_count()
        scripts = self._script_count()
        stylesheets = self._stylesheet_count()
        height = self.height

        complexity = (images * 10) + (scripts / 5) + (stylesheets / 10)
        score = complexity * (height * 10)
        score = round(score / 8000000)  # arbitrary number

        if score > Limits.RESIZE_MAX_WAIT_TIME:
            score = Limits.RESIZE_MAX_WAIT_TIME

        # The score can generally be thought of as the number
        # of seconds to wait
        while score > 0:
            time.sleep(Limits.RESIZE_WAIT_TIME)
            score -= 1


class Addons:
    """Firefox addons."""

    REFERER_HEADER = pkg_resources.resource_filename(
        'extensions',
        'referer_header.xpi'
    )

    # https://github.com/gorhill/uBlock/releases
    UBLOCK_ORIGIN = pkg_resources.resource_filename(
        'extensions',
        'uBlock0_1.17.7rc0.xpi'
    )

    # https://www.i-dont-care-about-cookies.eu/
    IDCAC = pkg_resources.resource_filename(
        'extensions',
        'idcac_2.9.5.xpi'
    )


class Limits:
    """Limits class.

    Various limits used in rendering page.
    """

    ALL_IMAGES_LOADED = 15

    SLEEP_BETWEEN_IMAGE_LOAD_CHECK = 1

    RESIZE_WAIT_TIME = 1

    RESIZE_MAX_WAIT_TIME = 25


class UserAgents:
    """User agents."""

    # masquerading as google bot can trick some website to not serve
    # tons of ads and load faster
    GOOGLEBOT = 'Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)'



"""Crawler test."""

from saas.storage.index import Index, EmptySearchResultException
import saas.utils.console as console
from saas.crawler.crawler import Crawler
from os.path import dirname, realpath
from unittest.mock import MagicMock
from saas.web.url import Url
import unittest
import os


class TestCrawler(unittest.TestCase):
    """Test crawler class."""

    def setUp(self):
        """Set up test."""
        self.console = console
        self.path_to_url_source = dirname(realpath(__file__)) + '/urls'
        open(self.path_to_url_source, 'w+').close()

    def tearDown(self):
        """Tear down test."""
        self.crawler.stop()
        os.remove(self.path_to_url_source)

    def add_url_source(self, url: str):
        """Add a url to source list.

        Args:
            url: url string
        """
        f = open(self.path_to_url_source, 'a')
        f.write(url + '\n')
        f.close()

    def test_crawler_can_read_next_url_from_source(self):
        """Test crawler can read next url from source."""
        self.add_url_source('https://example.com')

        self.crawler = Crawler(self.path_to_url_source, Index())
        self.assertEqual(
            Url.from_string('https://example.com').to_string(),
            self.crawler._next_url().to_string()
        )

    def test_crawler_removes_urls_read_from_source(self):
        """Test crawler removes urls read from source."""
        self.add_url_source('https://example.com')
        self.add_url_source('https://example.com/foo')
        self.add_url_source('https://example.com/bar')

        self.crawler = Crawler(self.path_to_url_source, Index())

        # first line should now be https://example.com
        self.assertEqual(
            Url.from_string('https://example.com').to_string(),
            self.crawler._next_url().to_string()
        )

        # first line should now be https://example.com/foo
        self.assertEqual(
            Url.from_string('https://example.com/foo').to_string(),
            self.crawler._next_url().to_string()
        )

        # first line should now be https://example.com/bar
        self.assertEqual(
            Url.from_string('https://example.com/bar').to_string(),
            self.crawler._next_url().to_string()
        )

        self.crawler = Crawler(self.path_to_url_source, Index())

    def test_crawler_can_read_next_url_from_index(self):
        """Test crawler can read next url from source."""
        index = Index()
        url = Url.from_string('https://example.com/foo')
        index.remove_uncrawled_url = MagicMock()
        index.random_uncrawled_url = MagicMock(return_value=url)

        self.crawler = Crawler(self.path_to_url_source, index)

        self.assertEqual(
            Url.from_string('https://example.com/foo').to_string(),
            self.crawler._next_url().to_string()
        )
        index.remove_uncrawled_url.assert_called_with(url.hash())

    def test_next_url_returns_none_if_no_url_was_found(self):
        """Test _next_url() returns None if no url was found."""
        index = Index()
        index.random_uncrawled_url = MagicMock()
        index.random_uncrawled_url.side_effect = EmptySearchResultException()
        self.crawler = Crawler(self.path_to_url_source, index)

        self.assertEqual(
            None,
            self.crawler._next_url()
        )


if __name__ == '__main__':
    unittest.main()



"""Filesystem module."""

from __future__ import annotations
from saas.mount.file import Path, Directory, File, LastCapture
from saas.storage.index import Index, PhotoNotFoundException
from saas.storage.refresh import RefreshRate
from saas.utils.files import real_path
import saas.utils.console as console
from typing import Type, Generator
from fuse import FUSE, Operations
import errno
import os


def mount(mountpoint: str, index: Index, refresh_rate: Type[RefreshRate]):
    """Mount filesystem.

    Mount filesystem at given path.

    Args:
        mountpoint: where to mount filesystem
        index: index to read data from
        refresh_rate: Which refresh rate filesystem should use
            for fetching photos
    """
    filesystem = Filesystem(index, refresh_rate)
    FUSE(filesystem, real_path(mountpoint), nothreads=True, foreground=True)


class Filesystem(Operations):
    """Filesystem.

    The filesystem class can respond to requests for files
    from the kernel using FUSE

    See:
        https://github.com/libfuse/libfuse
    """

    ROOT_PATH = '/'

    def __init__(self, index: Index, refresh_rate: Type[RefreshRate]):
        """Create new filesystem.

        Args:
            index: Index where photos are stored
            refresh_rate: Which refresh rate filesystem should use
                for fetching photos
        """
        self.index = index
        self.refresh_rate = refresh_rate

    def getattr(self, path: str, fh=None) -> dict:
        """Get attributes of file.

        Args:
            path: path to file
            fh: file descriptor

        Returns:
            Dictionary with file attributes
            dict

        Raises:
            FileNotFoundError: If a file was not found at path
        """
        console.df(f'getattr {path}')
        try:
            return self._attributes(path)
        except PhotoNotFoundException:
            pass
        except FileNotFoundError:
            pass
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)

    def readdir(self, path: str, fh: int) -> Generator:
        """Read directory.

        List directory contents.

        Args:
            path: path to directory
            fh: file descriptor

        Yields:
            File in directory
            str
        """
        console.df(f'readdir {path}')
        for file in self._list(path):
            yield file.filename

    def open(self, path: str, flags: int) -> int:
        """Open file for low level io.

        Args:
            path: path to file
            flags: flags

        Returns:
            Returns a file descriptor id
            int
        """
        console.df(f'open {path}')
        return os.open(self._translate_path(path), flags)

    def release(self, path: str, fh: int):
        """Release file.

        Closes file descriptor, so that it no longer
        refers to any file and may be reused.

        Args:
            path: path to a file
            fh: file descriptor
        """
        console.df(f'release {path}')
        os.close(fh)

    def read(self, path: str, length: int, offset: int, fh: int) -> bytes:
        """Read from file.

        Args:
            path: path to file
            length: number of bytes to read
            offset: where to start in file
            fh: file descriptor

        Returns:
            Contents of file
            bytes
        """
        console.df(f'read {path}')
        os.lseek(fh, offset, os.SEEK_SET)
        return os.read(fh, length)

    def write(self, path: str, data: str, offset: int, fh: int):
        """Write to file.

        Disabled.

        Args:
            path: path to file
            data: data to write
            length: number of bytes to read
            offset: where to start in file

        Raises:
            IOError: will always raise this exception
        """
        console.df(f'write {path}')
        raise IOError(
            errno.EPERM,
            os.strerror(errno.EPERM),
            path
        )

    def _attributes(self, path: str) -> dict:
        """Get attributes of file at path.

        Args:
            path: path to file

        Returns:
            Dictionary with file attributes, see files module
            dict

        Raises:
            FileNotFoundError: if no file exists at given path
        """
        if path == Filesystem.ROOT_PATH:
            return Directory.attributes()

        parsed = Path(path)

        if parsed.includes_domain() and not parsed.includes_captured_at():
            domains = self.index.photos_unique_domains(self.refresh_rate)
            if parsed.domain not in domains:
                raise FileNotFoundError(f'Unkown domain: {parsed.domain}')
            return Directory.attributes()

        if parsed.includes_captured_at() and not parsed.includes_end():
            captures = self.index.photos_unique_captures_of_domain(
                parsed.domain,
                self.refresh_rate
            )
            captures.append(LastCapture.FILENAME)
            if parsed.captured_at not in captures:
                raise FileNotFoundError(
                    f'Unkown capture: {parsed.captured_at}'
                )
            return Directory.attributes()

        if self.index.photos_directory_exists(
            domain=parsed.domain,
            captured_at=parsed.captured_at,
            directory=parsed.end_as_directory(),
            refresh_rate=self.refresh_rate
        ):
            return Directory.attributes()

        file_exists = self.index.photos_file_exists(
            domain=parsed.domain,
            captured_at=parsed.captured_at,
            full_filename=parsed.end_as_file(),
            refresh_rate=self.refresh_rate
        )
        if file_exists:
            return File.attributes(None, file_exists)

        raise FileNotFoundError(f'No file at path: {path}')

    def _list(self, path: str) -> list:
        """List directory.

        Args:
            path: path to directory

        Returns:
            List of directory content
            list
        """
        if path == Filesystem.ROOT_PATH:
            return self._list_root()

        parsed = Path(path)

        if not parsed.includes_captured_at():
            return self._list_unique_captures(parsed.domain)

        return self._list_directory(
            parsed.domain,
            parsed.captured_at,
            parsed.end_as_directory()
        )

    def _list_root(self) -> list:
        """List root directory.

        Returns:
            List of domains
            list
        """
        files = self._current_and_parent_dirs()
        for domain in self.index.photos_unique_domains(self.refresh_rate):
            files.append(Directory(domain))
        return files

    def _list_unique_captures(self, domain: str) -> list:
        """List unique captures of domain.

        Args:
            domain: domain name to list

        Returns:
            List of captures.
            list
        """
        files = self._current_and_parent_dirs()

        captures = self.index.photos_unique_captures_of_domain(
            domain,
            self.refresh_rate
        )
        for file in captures:
            files.append(Directory(file))

        files.append(Directory(LastCapture.FILENAME))

        return files

    def _list_directory(
        self,
        domain: str,
        captured_at: str,
        directory: str
    ) -> list:
        """List contents of a directory.

        Args:
            domain: filter by domain
            captured_at: filter by capture time
            directory: directory path

        Returns:
            A list of files and directories
            list
        """
        files = self._current_and_parent_dirs()

        result = self.index.photos_list_files_in_directory(
            domain,
            captured_at,
            directory,
            self.refresh_rate
        )
        for file in result:
            files.append(File(file))

        result = self.index.photos_list_directories_in_directory(
            domain,
            captured_at,
            directory,
            self.refresh_rate
        )
        for file in result:
            files.append(Directory(file))

        return files

    def _current_and_parent_dirs(self) -> list:
        """Get current and parent directory.

        Returns:
            List of directories
            list
        """
        return [
            Directory('.'),
            Directory('..'),
        ]

    def _translate_path(self, path: str) -> str:
        """Translate path to path in data dir.

        Args:
            path: Path in mounted directory eg.
                /example.com/2019-01-13H20:00/foo/bar.png

        Returns:
            Path in data directory to raw data
            str

        Raises:
            FileNotFoundError: If file at given path does not exist
        """
        parsed = Path(path)
        exists = self.index.photos_file_exists(
            domain=parsed.domain,
            captured_at=parsed.captured_at,
            full_filename=parsed.end_as_file(),
            refresh_rate=self.refresh_rate
        )
        if not exists:
            raise FileNotFoundError('photo do not exist in index')

        photo = self.index.photos_get_photo(
            domain=parsed.domain,
            captured_at=parsed.captured_at,
            full_filename=parsed.end_as_file(),
            refresh_rate=self.refresh_rate
        )

        return photo.path.full_path()



"""Stats module."""

from saas.storage.index import Index
import psutil
import time
import os


def throughput(index: Index, timeframe: int) -> int:
    """Get current throughput.

    Args:
        index: Index photos are stored in
        timeframe: timeframe in minutes

    Returns:
        number of photos stored in index during timeframe
        int
    """
    return index.calculate_throughput(timeframe)


def cpu_usage(sample: int) -> float:
    """Calculate average cpu usage.

    Waits 1s between samples.

    Args:
        sample: number of times to measure

    Returns:
        average cpu usage
        float
    """
    samples = []
    while sample > 0:
        sample -= 1
        samples.append(psutil.cpu_percent())
        time.sleep(1)
    return round(sum(samples) / len(samples), 1)


def memory_usage(sample: int) -> float:
    """Calculate average memory usage.

    Waits 1s between samples.

    Args:
        sample: number of times to measure

    Returns:
        average memory usage
        float
    """
    samples = []
    while sample > 0:
        sample -= 1
        samples.append(psutil.virtual_memory().percent)
        time.sleep(1)
    return round(sum(samples) / len(samples), 1)


def load_avg(timeframe: int) -> float:
    """Load average.

    Args:
        timeframe: last 1min, 5mins or 15mins

    Returns:
        cpu load
        float

    Raises:
        ValueError: if timeframe is unsupported
    """
    if timeframe == 1:
        return round(os.getloadavg()[0], 2)

    if timeframe == 5:
        return round(os.getloadavg()[1], 2)

    if timeframe == 15:
        return round(os.getloadavg()[2], 2)

    raise ValueError(f'unsupported timeframe {timeframe}')



"""saas entry point."""

from saas.storage.index import Index, EmptySearchResultException
from saas.photographer.javascript import JavascriptSnippets
from saas.storage.datadir import DataDirectory
import saas.storage.refresh as refresh
import saas.utils.console as console
import saas.utils.args as arguments
from saas.threads import Controller
import time
import sys


def main():
    """Entry point for saas."""
    try:

        parser = arguments.get_argument_parser()
        args = parser.parse_args(sys.argv[1:])

        console.DEBUG = args.debug

        JavascriptSnippets.load()

        index = Index(host=args.elasticsearch_host)

        if not index.ping():
            console.p('ERROR: failed to connect to elasticsearch')
            sys.exit()

        if not index.verify():
            if not args.setup_elasticsearch and not args.clear_elasticsearch:
                console.p('ERROR: elasticsearch is not configured')
                console.p('       {} {}'.format(
                    'start saas with --setup-elasticsearch',
                    'to configure elasticsearch'
                ))
                sys.exit()

        datadir = DataDirectory(args.data_dir, args.optimize_storage)

        refresh_rate = {
            'day': refresh.Daily,
            'hour': refresh.Hourly,
            'minute': refresh.EveryMinute,
        }[args.refresh_rate]

        if args.setup_elasticsearch:
            index.create_indices()

        if args.clear_elasticsearch:
            index.clear()
            index.create_indices()

        if args.clear_data_dir:
            datadir.clear()

        if not Controller.start_filesystem(
            mountpoint=args.mountpoint,
            datadir=datadir,
            refresh_rate=refresh_rate,
            elasticsearch_host=args.elasticsearch_host
        ):
            sys.exit()

        Controller.start_stats(
            elasticsearch_host=args.elasticsearch_host
        )

        Controller.start_crawlers(
            amount=args.crawler_threads,
            url_file=args.url_file,
            ignore_found_urls=args.ignore_found_urls,
            stay_at_domain=args.stay_at_domain,
            elasticsearch_host=args.elasticsearch_host,
            debug=args.debug
        )

        Controller.start_photographers(
            amount=args.photographer_threads,
            refresh_rate=refresh_rate,
            datadir=datadir,
            viewport_width=args.viewport_width,
            viewport_height=args.viewport_height,
            viewport_max_height=args.viewport_max_height,
            elasticsearch_host=args.elasticsearch_host,
            debug=args.debug
        )

        while True:

            if args.stop_if_idle == 0:
                time.sleep(10)
                continue

            try:
                crawled = index.timestamp_of_most_recent_document(
                    index.CRAWLED
                )
                photos = index.timestamp_of_most_recent_document(
                    index.PHOTOS
                )

                timestamp = photos
                if crawled > timestamp:
                    timestamp = crawled

                seconds = int(time.time()) - timestamp
                mins = int(seconds / 60)
                if mins >= args.stop_if_idle:
                    console.p(f'was idle for {mins} minutes', end='')
                    raise StopIfIdleTimeoutExpired

            except EmptySearchResultException:
                pass
            finally:
                time.sleep(2)

    except (KeyboardInterrupt, StopIfIdleTimeoutExpired):
        console.p(' terminating.')
        Controller.stop_all()
        console.p('')


class StopIfIdleTimeoutExpired(Exception):
    """Idle timeout expired."""

    pass


if __name__ == '__main__':
    main()



"""Storage module."""

from __future__ import annotations
from saas.photographer.photo import Photo, PhotoPath, Screenshot
from elasticsearch.exceptions import RequestError
from saas.storage.datadir import DataDirectory
from saas.storage.refresh import RefreshRate
from urllib.error import HTTPError, URLError
from datetime import datetime, timedelta
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk
import saas.utils.console as console
from saas.web.url import Url, UrlId
import saas.mount.file as file
from typing import Type
import urllib.request
import random
import json
import time


class Index:
    """Index storage.

    Wrapper around elasticsearch api
    """

    UNCRAWLED = 'uncrawled'

    CRAWLED = 'crawled'

    PHOTOS = 'photos'

    def __init__(
        self,
        datadir: DataDirectory=None,
        es_client: Elasticsearch=None,
        host: str='localhost:9200'
    ):
        """Create new index.

        Args:
            datadir: Data directory (default: {None})
            es_client: Elasticsearch client,
                useful in testing (default: {None})
            host: elasticsearch host (default: {'localhost:9200'})
        """
        self.datadir = datadir
        self.host = host
        if es_client is not None:
            self.es = es_client
        else:
            self.es = Elasticsearch(
                [self.host],
                max_retries=2,
                retry_on_timeout=True
            )

    def ping(self) -> bool:
        """Ping elastic search server.

        Returns:
            returns True if elasticsearch responded, otherwise False
            bool
        """
        try:
            with urllib.request.urlopen(f'http://{self.host}') as response:
                response.read()
                return True
        except (HTTPError, URLError):
            pass
        return False

    def verify(self):
        """Verify elasticsearch is configured properly.

        Returns:
            True if configured correctly, otherwise False
            bool
        """
        url = f'http://{self.host}/_mapping'
        with urllib.request.urlopen(url) as response:
            mappings = json.loads(response.read())
            if Index.UNCRAWLED not in mappings:
                return False
            if Index.CRAWLED not in mappings:
                return False
            if Index.PHOTOS not in mappings:
                return False
        return True

    def clear(self):
        """Clear all documents."""
        console.p('clearing all indices')
        self.es.indices.delete(index='_all', request_timeout=1000000)
        console.p('indices cleared')

    def create_indices(self):
        """Create indices in elasticsearch."""
        console.p('creating indices')
        try:
            self.es.indices.create(Index.UNCRAWLED, body={
                'mappings': Mappings.uncrawled
            })
            self.es.indices.create(Index.CRAWLED, body={
                'mappings': Mappings.crawled
            })
            self.es.indices.create(Index.PHOTOS, body={
                'mappings': Mappings.photos,
                'settings': Settings.photos,
            })
            console.p('done.')
        except RequestError:
            console.p('indices already exist, skipping.')

    def calculate_throughput(self, timeframe: int) -> int:
        """Calculate throughput.

        Number of photos stored in index during timeframe

        Args:
            timeframe: timeframe in minutes

        Returns:
            number of photos stored in index during timeframe
            int
        """
        now = datetime.now()
        start = int((now - timedelta(minutes=timeframe)).timestamp())
        end = now.timestamp()

        res = self.es.search(index=Index.PHOTOS, size=0, body={
            'query': {
                'range': {
                    'timestamp': {
                        'gte': start,
                        'lte': end,
                    },
                }
            }
        })
        throughput = res['hits']['total']  # type: int
        return throughput

    def add_crawled_url(self, url: Url):
        """Add crawled url.

        Args:
            url: A url that have been crawled
        """
        self.add_crawled_urls([url])

    def add_crawled_urls(self, urls: list):
        """Add crawled urls.

        Args:
            urls: A list of urls that have been crawled
        """
        prepared = self._prepare_urls(urls, Index.CRAWLED)
        bulk(self.es, prepared, request_timeout=80)

    def timestamp_of_most_recent_document(self, index: str) -> int:
        """Get timestamp of most recent document in given index.

        Args:
            index: the index the most recent document should be in

        Returns:
            Timestamp of most recent document in given index
            int

        Raises:
            EmptySearchResultException: if index is empty
        """
        res = self.es.search(index=index, size=1, body={
            'query': {
                'match_all': {}
            },
            'sort': [
                {
                    'timestamp': {
                        'order': 'desc'
                    }
                }
            ],
            '_source': ['timestamp']
        })

        if res['hits']['total'] == 0:
            raise EmptySearchResultException('index is empty')

        doc = res['hits']['hits'][0]  # type: dict
        timestamp = doc['_source']['timestamp']  # type: int
        return timestamp

    def add_uncrawled_urls(self, urls: list):
        """Add uncrawled urls.

        Args:
            urls: A list of urls that have NOT been crawled yet
        """
        urls = self.remove_already_crawled_urls(urls)
        prepared = self._prepare_urls(urls, Index.UNCRAWLED)
        bulk(self.es, prepared, request_timeout=80)

    def remove_already_crawled_urls(self, urls: list) -> list:
        """Remove already crawled urls from a list of urls.

        Args:
            urls: A list of urls

        Returns:
            A cleaned list of uncrawled urls
            list
        """
        hashes = []
        for url in urls:
            hashes.append(url.hash())
        res = self.es.search(index=Index.CRAWLED, body={
            'query': {
                'bool': {
                    'filter': {
                        'terms': {
                            '_id': hashes
                        }
                    }
                }
            },
            'stored_fields': []
        })
        for doc in res['hits']['hits']:
            while doc['_id'] in hashes:
                hashes.remove(doc['_id'])
        out = []
        for url in urls:
            if url.hash() in hashes:
                out.append(url)
        return out

    def _prepare_urls(self, urls: list, index: str) -> list:
        """Prepare urls for bulk add.

        Args:
            urls: list of Urls
            index: index to add urls to

        Returns:
            Prepared list of urls
            list
        """
        prepared = []
        for url in urls:
            if index == Index.CRAWLED:
                source = {
                    'url': url.to_string(),
                    'timestamp': int(time.time()),
                    'lock_value': '',
                    'lock_format': ''
                }
            else:
                source = {
                    'url': url.to_string(),
                    'timestamp': int(time.time()),
                }
            prepared.append({
                '_type': 'url',
                '_index': index,
                '_id': url.hash(),
                '_source': source
            })
        return prepared

    def random_uncrawled_url(self) -> Url:
        """Get random uncrawled url.

        Returns:
            An uncrawled url
            Url

        Raises:
            EmptySearchResultException: if index is empty
        """
        res = self.es.search(index=Index.UNCRAWLED, size=1, body={
            'query': {
                'function_score': {
                    'query': {
                        'bool': {
                            'must_not': {
                                'term': {
                                    'crawled': True
                                }
                            }
                        }
                    },
                    'random_score': {}
                }
            }
        })

        if res['hits']['total'] == 0:
            raise EmptySearchResultException('uncrawled index is empty')

        url = Url.from_string(res['hits']['hits'][0]['_source']['url'])
        return url

    def recently_crawled_url(self, refresh_rate=RefreshRate):
        """Get recently crawled url.

        Picks from the last 5 crawled to prevent two running
        photograpers to fetch the same one.

        Args:
            refresh_rate: the refresh reate to search for and
                use to avoid locked urls

        Returns:
            A url that has been crawled with status code 200
            Url

        Raises:
            EmptySearchResultException: if no url was found
        """
        res = self.es.search(index=Index.CRAWLED, size=5, body={
            'query': {
                'bool': {
                    'must': {
                        'term': {
                            'status_code': 200,
                        }
                    },
                    'must_not': [
                        {
                            'term': {
                                'lock_value': refresh_rate().lock(),
                            }
                        }
                    ]
                }
            },
            'sort': [
                {
                    'timestamp': {
                        'order': 'desc'
                    }
                }
            ]
        })

        if res['hits']['total'] == 0:
            raise EmptySearchResultException('no crawled url was found')

        hits = res['hits']['hits']

        return Url.from_string(random.choice(hits)['_source']['url'])

    def crawled_urls_count(self, refresh_rate=RefreshRate) -> int:
        """Crawled url count.

        Args:
            refresh_rate: the refresh reate to search for and
                use to avoid locked urls

        Returns:
            number of crawled urls with status code 200
            int
        """
        res = self.es.search(index=Index.CRAWLED, size=0, body={
            'query': {
                'bool': {
                    'must': {
                        'term': {
                            'status_code': 200,
                        }
                    },
                    'must_not': [
                        {
                            'term': {
                                'lock_value': refresh_rate().lock(),
                            }
                        }
                    ]
                }
            },
            'sort': [
                {
                    'timestamp': {
                        'order': 'desc'
                    }
                }
            ]
        })

        count = res['hits']['total']  # type: int
        return count

    def remove_uncrawled_url(self, id: str):
        """Remove url from uncrawled index.

        Args:
            id: Id of url to delete
        """
        self.es.delete(
            index=Index.UNCRAWLED,
            doc_type='url',
            id=id,
            ignore=404
        )

    def set_status_code_for_crawled_url(self, url: Url, status_code: int):
        """Set status code for a crawled url.

        Args:
            url: Url to set status code of
            status_code: the status code of the http request to the url
        """
        self.es.update(
            index=Index.CRAWLED,
            doc_type='url',
            id=url.hash(),
            retry_on_conflict=3,
            body={
                'doc': {
                    'status_code': status_code
                }
            }
        )

    def lock_crawled_url(self, url: Url, refresh_rate: Type[RefreshRate]):
        """Lock a crawld url.

        Place a lock on a crawled url for a given refresh rate.

        Args:
            url: Url to lock
            refresh_rate: Refresh rate to use (Hourly, Daily, etc.)
        """
        self.es.update(
            index=Index.CRAWLED,
            doc_type='url',
            id=url.hash(),
            retry_on_conflict=3,
            body={
                'doc': {
                    'lock_format': refresh_rate.lock_format(),
                    'lock_value': refresh_rate().lock(),
                }
            }
        )

    def save_photo(self, photo: Photo):
        """Save photo in index.

        Will not store the actual photo data, this should be stored
        in the data directory.

        Args:
            photo: Photo to store
        """
        self.es.index(
            index=Index.PHOTOS,
            doc_type='photo',
            id=photo.path.uuid,
            body={
                'url_id': photo.url.hash(),
                'refresh_rate': photo.refresh_rate.lock_format(),
                'captured_at': photo.refresh_rate().lock(),
                'filesize': photo.filesize(),
                'filename': photo.filename(),
                'directory': photo.directory(),
                'domain': photo.domain(),
                'timestamp': int(time.time())
            }
        )

    def photos_unique_domains(self, refresh_rate: Type[RefreshRate]) -> list:
        """Get unique domains that pictures have been taken of.

        Args:
            refresh_rate: Given refresh rate photo was taken with
        """
        res = self.es.search(index=Index.PHOTOS, size=0, body={
            'query': {
                'bool': {
                    'must': {
                        'term': {
                            'refresh_rate': refresh_rate.lock_format(),
                        }
                    },
                }
            },
            'aggs': {
                'photos': {
                    'terms': {
                        'field': 'domain',
                        'size': 10000
                    }
                }
            }
        })
        unique = []
        for bucket in res['aggregations']['photos']['buckets']:
            unique.append(bucket['key'])
        return unique

    def photos_unique_captures_of_domain(
        self,
        domain: str,
        refresh_rate: Type[RefreshRate]
    ) -> list:
        """Get unique captures for a domain.

        Args:
            domain: The given domain to check
            refresh_rate: Given refresh rate photo was taken with
        """
        res = self.es.search(index=Index.PHOTOS, size=0, body={
            'query': {
                'bool': {
                    'must': [
                        {
                            'term': {
                                'domain': domain,
                            }
                        },
                        {
                            'term': {
                                'refresh_rate': refresh_rate.lock_format(),
                            }
                        },
                    ],
                }
            },
            'aggs': {
                'photos': {
                    'terms': {
                        'field': 'captured_at',
                        'size': 10000
                    }
                }
            }
        })
        unique = []
        for bucket in res['aggregations']['photos']['buckets']:
            unique.append(bucket['key'])
        return unique

    def photos_most_recent_capture_of_domain(
        self,
        domain: str,
        refresh_rate: Type[RefreshRate]
    ) -> str:
        """Get most recently captured_at value of domain.

        Args:
            domain: domain to check
            refresh_rate: Given refresh rate photo was taken with

        Returns:
            The most recent capture_at value in photos index
            str

        Raises:
            EmptySearchResultException: if no capture_at value was found
        """
        res = self.es.search(index=Index.PHOTOS, size=1, body={
            'query': {
                'bool': {
                    'must': [
                        {
                            'term': {
                                'domain': domain
                            }
                        },
                        {
                            'term': {
                                'refresh_rate': refresh_rate.lock_format()
                            }
                        }
                    ]
                }
            },
            'sort': [
                {
                    'timestamp': {
                        'order': 'desc'
                    }
                }
            ]
        })

        if res['hits']['total'] == 0:
            raise EmptySearchResultException(
                f'no capture for given refresh rate and {domain} was found'
            )

        doc = res['hits']['hits'][0]  # type: dict
        captured_at = doc['_source']['captured_at']  # type: str
        return captured_at

    def photos_get_photo(
        self,
        domain: str,
        captured_at: str,
        full_filename: str,
        refresh_rate: Type[RefreshRate]
    ) -> Photo:
        """Get photo from photos index.

        Args:
            domain: domain photo belongs to
            captured_at: when photo was captured
            full_filename: full filename of photo
                eg. /some/path/some-filename.png
            refresh_rate: Given refresh rate photo was taken with

        Returns:
            Requested photo metadata
            Photo

        Raises:
            PhotoNotFoundException: If photo was not found
        """
        directory = '/'.join(full_filename.split('/')[:-1])
        directory = directory.rstrip('/') + '/'
        filename = full_filename.split('/')[-1:][0]

        captured_at = file.LastCapture.translate(
            captured_at,
            domain,
            self,
            refresh_rate
        )

        res = self.es.search(index=Index.PHOTOS, size=1, body={
            'query': {
                'bool': {
                    'must': [
                        {
                            'term': {
                                'domain': domain
                            }
                        },
                        {
                            'term': {
                                'refresh_rate': refresh_rate.lock_format()
                            }
                        },
                        {
                            'term': {
                                'captured_at': captured_at
                            }
                        },
                        {
                            'term': {
                                'directory': directory
                            }
                        },
                        {
                            'term': {
                                'filename': filename
                            }
                        }
                    ]
                }
            },
        })

        if res['hits']['total'] == 0:
            raise PhotoNotFoundException('no photo was found')

        res = res['hits']['hits'][0]
        uuid = res['_id']

        if self.datadir is None:
            raise Exception('Cannot get photo from Index without a data dir')

        path = PhotoPath(self.datadir, uuid=uuid)

        photo = Screenshot(
            url=UrlId(res['_source']['url_id']),
            path=path,
            refresh_rate=refresh_rate,
            index_filesize=res['_source']['filesize']
        )

        return photo

    def photos_file_exists(
        self,
        domain: str,
        captured_at: str,
        full_filename: str,
        refresh_rate: Type[RefreshRate]
    ):
        """Check if photo exists in photos index.

        Args:
            domain: domain photo belongs to
            captured_at: when photo was captured
            full_filename: full filename of photo
                eg. /some/path/some-filename.png
            refresh_rate: Given refresh rate photo was taken with

        Returns:
            False if file was not found, if found it's filesize
            is returned
            bool or int
        """
        captured_at = file.LastCapture.translate(
            captured_at,
            domain,
            self,
            refresh_rate
        )
        try:
            photo = self.photos_get_photo(
                domain,
                captured_at,
                full_filename,
                refresh_rate
            )
            return photo.filesize()
        except PhotoNotFoundException:
            return False

    def photos_directory_exists(
        self,
        domain: str,
        captured_at: str,
        directory: str,
        refresh_rate: Type[RefreshRate]
    ) -> bool:
        """Check if directory exists in photos index.

        Args:
            domain: domain photo belongs to
            captured_at: when photo was captured
            directory: directory path eg. /some/path/to/dir/
            refresh_rate: Given refresh rate photo was taken with

        Returns:
            True if photo was found, else False
            bool
        """
        captured_at = file.LastCapture.translate(
            captured_at,
            domain,
            self,
            refresh_rate
        )
        res = self.es.search(index=Index.PHOTOS, size=0, body={
            'query': {
                'bool': {
                    'must': [
                        {
                            'term': {
                                'domain': domain,
                            }
                        },
                        {
                            'term': {
                                'refresh_rate': refresh_rate.lock_format(),
                            }
                        },
                        {
                            'term': {
                                'captured_at': captured_at,
                            }
                        },
                        {
                            'wildcard': {
                                'directory': directory + '*',
                            }
                        }
                    ],
                }
            },
            'aggs': {
                'photos': {
                    'terms': {
                        'field': 'directory',
                        'size': 10000
                    }
                }
            }
        })
        for bucket in res['aggregations']['photos']['buckets']:
            if directory in bucket['key']:
                return True
        return False

    def photos_list_files_in_directory(
        self,
        domain: str,
        captured_at: str,
        directory: str,
        refresh_rate: Type[RefreshRate]
    ) -> list:
        """List photos in a directory.

        Args:
            domain: domain photos should belong to
            captured_at: when photos should have been captured
            directory: directory files should be located in
            refresh_rate: Given refresh rate photo was taken with

        Returns:
            A list of files
            list
        """
        captured_at = file.LastCapture.translate(
            captured_at,
            domain,
            self,
            refresh_rate
        )
        res = self.es.search(index=Index.PHOTOS, size=10000, body={
            'query': {
                'bool': {
                    'must': [
                        {
                            'term': {
                                'domain': domain
                            }
                        },
                        {
                            'term': {
                                'refresh_rate': refresh_rate.lock_format()
                            }
                        },
                        {
                            'term': {
                                'captured_at': captured_at
                            }
                        },
                        {
                            'term': {
                                'directory': directory
                            }
                        }
                    ]
                }
            }
        })
        files = []
        for doc in res['hits']['hits']:
            if doc['_source']['filesize'] < 100:
                files.append(
                    doc['_source']['filename'] + file.Path.RENDERING_EXTENSION
                )
            else:
                files.append(doc['_source']['filename'])
        return files

    def photos_list_directories_in_directory(
        self,
        domain: str,
        captured_at: str,
        directory: str,
        refresh_rate: Type[RefreshRate]
    ) -> list:
        """List directories in a directory.

        Args:
            domain: domain directory should belong to
            captured_at: when directory should have been captured
            directory: directory should be located in
            refresh_rate: given refresh rate photo was taken with

        Returns:
            A list of directories
            list
        """
        captured_at = file.LastCapture.translate(
            captured_at,
            domain,
            self,
            refresh_rate
        )
        res = self.es.search(index=Index.PHOTOS, size=0, body={
            'query': {
                'bool': {
                    'must': [
                        {
                            'term': {
                                'domain': domain,
                            }
                        },
                        {
                            'term': {
                                'refresh_rate': refresh_rate.lock_format(),
                            }
                        },
                        {
                            'term': {
                                'captured_at': captured_at,
                            }
                        },
                        {
                            'wildcard': {
                                'directory': directory + '*',
                            }
                        }
                    ],
                }
            },
            'aggs': {
                'photos': {
                    'terms': {
                        'field': 'directory',
                        'size': 10000
                    }
                }
            }
        })
        directories = []  # type: list
        for bucket in res['aggregations']['photos']['buckets']:
            # trim the parent directories from the path,
            # and strip it's child directories
            path = bucket['key']
            path = path.replace(directory, '', 1)
            path = path.split('/')[0]
            if path not in directories and path != '':
                directories.append(path)
        return directories


class PhotoNotFoundException(Exception):
    """Photo was not found exception."""

    pass


class Mappings:
    """Mappings for elasticsearch indices."""

    uncrawled = {
        'url': {
            'properties': {
                'url': {
                    'type': 'text',
                    'fields': {
                        'keyword': {
                            'type': 'keyword',
                            'ignore_above': 256
                        }
                    }
                },
                'timestamp': {
                    'type': 'date',
                    'format': 'epoch_second',
                },
                'status_code': {
                    'type': 'short'
                }
            }
        }
    }

    crawled = {
        'url': {
            'properties': {
                'url': {
                    'type': 'text',
                    'fields': {
                        'keyword': {
                            'type': 'keyword',
                            'ignore_above': 256
                        }
                    }
                },
                'timestamp': {
                    'type': 'date',
                    'format': 'epoch_second',
                },
                'lock_format': {
                    'type': 'text'
                },
                'lock_value': {
                    'type': 'text'
                }
            }
        }
    }

    photos = {
        'photo': {
            'properties': {
                'url_id': {
                    'type': 'text',
                },
                'refresh_rate': {
                    'type': 'text'
                },
                'captured_at': {
                    'type': 'text',
                    'fielddata': True
                },
                'filesize': {
                    'type': 'integer',
                },
                'filename': {
                    'type': 'text',
                    'analyzer': 'analyzer_filename',
                },
                'directory': {
                    'type': 'text',
                    'analyzer': 'analyzer_directory',
                    'fielddata': True
                },
                'domain': {
                    'type': 'text',
                    'analyzer': 'analyzer_domain',
                    'fielddata': True
                },
                'timestamp': {
                    'type': 'date',
                    'format': 'epoch_second',
                }
            }
        }
    }


class Settings:
    """Settings for elasticsearch indices."""

    photos = {
        'index': {
            'analysis': {
                'analyzer': {
                    'analyzer_domain': {
                        'tokenizer': 'whitespace',
                        'filter': 'lowercase'
                    },
                    'analyzer_directory': {
                        'tokenizer': 'path_hierarchy'
                    },
                    'analyzer_filename': {
                        'tokenizer': 'whitespace',
                        'filter': 'lowercase'
                    }
                }
            }
        }
    }


class EmptySearchResultException(Exception):
    """Empty search result."""

    pass



"""Photo module test."""

from saas.photographer.photo import PhotoPath, LoadingPhoto
from saas.storage.datadir import DataDirectory
import saas.storage.refresh as refresh
from os.path import dirname, isfile
from saas.web.url import Url
import unittest


class TestPhoto(unittest.TestCase):
    """Test photo class."""

    def setUp(self):
        """Set up test."""
        self.datadir = DataDirectory(dirname(__file__) + '/datadir')

    def tearDown(self):
        """Tear down test."""
        self.datadir.remove_data_dir()

    def test_photo_path_can_be_created(self):
        """Pass."""
        path = PhotoPath(self.datadir)
        self.assertGreater(len(path.uuid), 20)
        self.assertIn(member=self.datadir.root, container=path.full_path())

    def test_loading_photo_can_be_saved_to_datadir(self):
        """Test a loading photo can be saved to data directory."""
        path = PhotoPath(self.datadir)
        url = Url.from_string('https://example.com')
        photo = LoadingPhoto(
            url=url,
            path=path,
            refresh_rate=refresh.Hourly
        )
        photo.save_loading_text()
        self.assertTrue(isfile(path.full_path()))
        self.assertEqual('loading', photo.get_raw())


if __name__ == '__main__':
    unittest.main()



"""Page module."""

from __future__ import annotations
from saas.web.url import Url


class Page:
    """Page class."""

    def __init__(self):
        """Create new page."""
        self.urls = []
        self.status_code = None
        self.content_type = ''

    def add_url(self, url: Url):
        """Add url.

        Args:
            url: Url to add
        """
        self.urls.append(url)

    def remove_urls_not_from_domain(self, domain: str):
        """Remove urls not from given domain.

        Args:
            domain: domain urls should be from
        """
        cleaned = []
        for url in self.urls:
            if url.domain == domain:
                cleaned.append(url)
        self.urls = cleaned



"""Threads module."""

from __future__ import annotations
from saas.crawler.crawler import Crawler, UrlFileNotFoundError
from saas.storage.datadir import DataDirectory
import saas.photographer.photographer as p
import saas.mount.filesystem as Filesystem
from saas.utils.files import real_path
import saas.storage.refresh as refresh
import saas.utils.console as console
from saas.storage.index import Index
from typing import Type, Optional
import saas.utils.stats as stats
from threading import Thread
import signal
import time
import uuid
import os


class Controller:
    """Thread Controller class."""

    SHOULD_RUN = True

    PHOTOGRAPHER_PROCESSES = 0

    FUSE_PID = None

    threads = {}  # type: dict

    webdrivers = []  # type: list

    @staticmethod
    def start_crawlers(
        amount: int,
        url_file: str,
        ignore_found_urls: bool,
        stay_at_domain: bool,
        elasticsearch_host: str,
        debug: bool
    ):
        """Start crawler threads.

        Args:
            amount: amount of crawlers to start
            url_file: path to urls file
            ignore_found_urls: if crawler should ignore new urls found on
                pages it crawls
            stay_at_domain: if crawler should ignore urls from a different
                domain than the one it was found at
            elasticsearch_host: elasticsearch host
            debug: Display debugging information
        """
        console.p(f'starting {amount} crawler threads')
        while amount > 0:
            thread_id = str(uuid.uuid4())
            thread = Thread(target=_crawler_thread, args=(
                url_file,
                ignore_found_urls,
                stay_at_domain,
                elasticsearch_host,
                debug,
                thread_id
            ))
            thread.start()
            Controller.threads[thread_id] = {
                'running': True
            }
            amount -= 1

    @staticmethod
    def start_stats(elasticsearch_host: str):
        """Start stats thread.

        Args:
            elasticsearch_host: elasticsearch host
        """
        thread = Thread(target=_stats_thread, args=(elasticsearch_host,))
        thread.start()

    @staticmethod
    def start_photographers(
        amount: int,
        refresh_rate: Type[refresh.RefreshRate],
        datadir: DataDirectory,
        viewport_width: int,
        viewport_height: int,
        viewport_max_height: Optional[int],
        elasticsearch_host: str,
        debug: bool
    ):
        """Start photographer threads.

        Args:
            amount: amount of crawlers to start
            refresh_rate: How often photographs should be refreshed,
                more exactly defines which lock should be placed on
                crawled urls
            datadir: Data directory to store pictures in
            viewport_width: width of camera viewport
            viewport_height: height of camera viewport
            viewport_max_height: max height of camera viewport
            elasticsearch_host: elasticsearch host
            debug: Display debugging information
        """
        console.p(f'starting {amount} photographer threads')
        Controller.PHOTOGRAPHER_PROCESSES = amount
        while amount > 0:
            thread_id = str(uuid.uuid4())
            thread = Thread(target=_photographer_thread, args=(
                refresh_rate,
                datadir,
                viewport_width,
                viewport_height,
                viewport_max_height,
                elasticsearch_host,
                debug,
                thread_id
            ))
            thread.start()
            Controller.threads[thread_id] = {
                'running': True
            }
            amount -= 1

    @staticmethod
    def start_filesystem(
        mountpoint: str,
        datadir: DataDirectory,
        refresh_rate: Type[refresh.RefreshRate],
        elasticsearch_host: str
    ):
        """Start filesystem process.

        FUSE python library will kill the main process,
        forking main process and mounts the filesystem
        from that process instead.

        Args:
            mountpoint: where to mount filesystem
            datadir: Data directory to store pictures in
            refresh_rate: Which refresh rate filesystem should use
                for fetching photos
            elasticsearch_host: elasticsearch host

        Returns:
            True if main process, False if the forked process
            bool
        """
        console.p(f'mounting filesystem at: {real_path(mountpoint)}')

        pid = os.fork()
        if pid != 0:
            Controller.FUSE_PID = pid
            return True

        try:
            Filesystem.mount(
                mountpoint,
                Index(datadir, host=elasticsearch_host),
                refresh_rate
            )
        except RuntimeError as e:
            console.p(f'failed to mount FUSE filesystem: {e}')

        return False

    @staticmethod
    def stop_all():
        """Stop all threads."""
        try:
            Controller.SHOULD_RUN = False

            i = 0
            while not Controller._any_thread_is_running():
                if i % 10 == 0:
                    console.p('waiting for saas to stop')
                i += 1
                time.sleep(0.5)

            console.p('cleaning up')
            try:
                for pid in Controller.webdrivers:
                    os.kill(pid, signal.SIGTERM)

                if Controller.FUSE_PID:
                    os.kill(Controller.FUSE_PID, signal.SIGTERM)
            except ProcessLookupError:
                pass
        except KeyboardInterrupt:
            Controller.stop_all()

    @staticmethod
    def _any_thread_is_running() -> bool:
        """Check if any thread is running.

        Returns:
            True if one or more thread is running, otherwise False
            bool
        """
        for thread_id in Controller.threads:
            if Controller.threads[thread_id]['running']:
                return False
        return True


def _crawler_thread(
    url_file: str,
    ignore_found_urls: bool,
    stay_at_domain: bool,
    elasticsearch_host: str,
    debug: bool,
    thread_id: str
):
    """Crawler thread.

    Args:
        url_file: path to url file
        ignore_found_urls: if crawler should ignore new urls found on
            pages it crawls
        stay_at_domain: if crawler should ignore urls from a different
            domain than the one it was found at
        elasticsearch_host: elasticsearch host
        debug: Display debugging information
        thread_id: id of thread
    """
    try:
        crawler = Crawler(
            url_file=url_file,
            index=Index(host=elasticsearch_host),
            ignore_found_urls=ignore_found_urls,
            stay_at_domain=stay_at_domain,
        )
        while Controller.SHOULD_RUN:
            crawler.tick()
    except UrlFileNotFoundError:
        console.p(f'ERROR: url_file was not found at \'{url_file}\'')
        time.sleep(2)
        Controller.threads[thread_id]['running'] = False
        Controller.stop_all()
    except Exception as e:
        console.p(f'error occured in crawler thread {thread_id}: {e}')
        if debug:
            raise e
    finally:
        Controller.threads[thread_id]['running'] = False


def _stats_thread(elasticsearch_host: str):
    """Stats thread.

    Prints system and saas statistics every 5th minute

    Args:
        elasticsearch_host: elasticsearch host
    """
    start = time.time()
    last_print = 1
    while Controller.SHOULD_RUN:

        time.sleep(1)
        mins = int(int(time.time() - start) / 60)
        if mins % 5 != 0 or mins <= last_print:
            continue

        index = Index(host=elasticsearch_host)
        last_print = mins

        t = '[throughput]           5m: {}, 15m: {}, 30min: {}, 1h: {}'.format(
            stats.throughput(index, 5),
            stats.throughput(index, 15),
            stats.throughput(index, 30),
            stats.throughput(index, 60),
        )
        ta = '{}  5m: {}, 15m: {}, 30min: {}, 1h: {}'.format(
            '[throughput 1min avg]',
            round(stats.throughput(index, 5) / 5, 2) if mins > 4 else 'n/a',
            round(stats.throughput(index, 15) / 15, 2) if mins > 14 else 'n/a',
            round(stats.throughput(index, 30) / 30, 2) if mins > 29 else 'n/a',
            round(stats.throughput(index, 60) / 60, 2) if mins > 59 else 'n/a',
        )
        load = '[load avg]             1m: {}, 5m: {}, 15min: {}'.format(
            stats.load_avg(1),
            stats.load_avg(5),
            stats.load_avg(15),
        )
        cpu = f'[current cpu usage]    {stats.cpu_usage(10)}%'
        mem = f'[memory usage]         {stats.memory_usage(10)}%'

        for msg in [t, ta, load, cpu, mem]:
            console.p(msg)


def _photographer_thread(
    refresh_rate: Type[refresh.RefreshRate],
    datadir: DataDirectory,
    viewport_width: int,
    viewport_height: int,
    viewport_max_height: Optional[int],
    elasticsearch_host: str,
    debug: bool,
    thread_id: str
):
    """Photographer thread.

    Args:
        refresh_rate: How often photographs should be refreshed
        datadir: Data directory to store pictures in
        viewport_width: width of camera viewport
        viewport_height: height of camera viewport
        viewport_max_height: max height of camera viewport
        elasticsearch_host: elasticsearch host
        debug: Display debugging information
        thread_id: id of thread
    """
    try:
        photographer = p.Photographer(
            Index(host=elasticsearch_host),
            refresh_rate,
            datadir,
            viewport_width,
            viewport_height,
            viewport_max_height
        )
        while Controller.SHOULD_RUN:
            photographer.tick()
    except Exception as e:
        console.p(f'error occured in photographer thread {thread_id}: {e}')
        if debug:
            raise e
    finally:
        Controller.threads[thread_id]['running'] = False



"""Url module."""

from __future__ import annotations
from urllib.parse import urlparse
import unicodedata
import hashlib
import re


class Url:
    """Url class."""

    def __init__(
            self,
            scheme: str,
            domain: str,
            path: str,
            query: str,
            fragment: str
    ):
        """Create new url.

        Args:
            scheme: eg. https
            domain: example.com
            path: /path/to/page.html
            query: ?some_param=value
            fragment: #some_string
        """
        self.scheme = scheme
        self.domain = domain
        self.path = path
        self.query = query
        self.fragment = fragment
        self.sha256 = ''

    @staticmethod
    def from_string(url: str) -> 'Url':
        """Create url from string.

        Args:
            url: string

        Returns:
            A url
            Url
        """
        parse = urlparse(url)
        if parse.scheme != 'http' and parse.scheme != 'https':
            raise InvalidUrlException('invalid url scheme')
        if '.' not in parse.netloc:
            raise InvalidUrlException('invalid domain scheme')
        path = parse.path
        while '//' in path:
            path = path.replace('//', '/')
        if len(path) > 0:
            if path[-1:] == '/':
                path = path[:-1]
        return Url(
            scheme=parse.scheme,
            domain=parse.netloc,
            path=path,
            query=parse.query,
            fragment=parse.fragment,
        )

    def to_string(self) -> str:
        """Convert url to string.

        Returns:
            The url as a string
            str
        """
        url = f'{self.scheme}://{self.domain}{self.path}'
        if self.query:
            url = f'{url}?{self.query}'
        if self.fragment:
            url = f'{url}#{self.fragment}'
        return url

    def hash(self) -> str:
        """Get sha256 hash of url.

        Returns:
            A sha256 hash of the url
            str
        """
        if self.sha256 != '':
            return self.sha256
        self.sha256 = hashlib.sha256(self.to_string().encode()).hexdigest()
        return self.sha256

    def make_filename(self) -> str:
        """Make filename.

        Make the filename used in filesystem.

        Returns:
            A safe filename to use in filesystem
            str
        """
        if self.fragment:
            with_fragment_filename = ''  # type: str
            with_fragment_filename = self._safe_filename(
                self.fragment
            ) + '.png'
            return with_fragment_filename

        if self.query:
            with_query_filename = ''  # type: str
            with_query_filename = self._safe_filename(self.query) + '.png'
            return with_query_filename

        filename = ''  # type: str
        filename = self._safe_filename(self.path.split('/')[-1:][0])
        if filename == '':
            filename = 'index'
        return filename + '.png'

    def make_directory(self) -> str:
        """Make directory.

        Make the relative path from data directory root directly
        above the result from make_filename().

        Returns:
            A path to url in data directory that is safe to use
            str
        """
        if self.path:
            dirs = self.path.split('/')
            for i, dir in enumerate(dirs):
                dirs[i] = self._safe_filename(dir)
            directory = '/'.join(dirs)
        else:
            directory = '/'

        if directory[-1:] != '/':
            directory += '/'

        if self.query:
            directory += '?/' + self._safe_filename(self.query) + '/'

        if self.fragment:
            directory += '#/' + self._safe_filename(self.fragment) + '/'

        directory = '/'.join(directory.split('/')[:-2])

        if directory[-1:] != '/':
            directory += '/'

        return directory

    def _safe_filename(self, value: str):
        """Make safe filename.

        Source:
            https://github.com/django/django/blob/master/django/utils/text.py
            basically copied the slugify function and added some handling
            of query params

        Args:
            value: A string value that might be unsafe for a filename

        Returns:
            A safe string to use as filename
            str
        """
        value = str(value)
        value = value.replace('=', '-')
        value = value.replace('&', '-and-')
        value = unicodedata.normalize(
            'NFKD',
            value
        ).encode('ascii', 'ignore').decode('ascii')
        value = re.sub(r'[^\w\s-]', '', value).strip().lower()
        return re.sub(r'[-\s]+', '-', value)

    def create_child_url(self, uri: str) -> 'Url':
        """Create child url from string.

        eg the child login of https://example.com with the uri
        /login would be https://example.com/login

        Args:
            uri: a path to a resource

        Returns:
            A fully qualified url
            Url
        """
        if len(uri) == 0:
            raise InvalidUrlException('uri was empty')
        if len(uri) > 1:
            if f'{uri[0]}{uri[1]}' == '//':
                return Url.from_string(f'{self.scheme}:{uri}')
        if uri[0] == '/':
            return Url.from_string(f'{self.scheme}://{self.domain}{uri}')
        if uri[0] == '#':
            return Url.from_string(f'{self.scheme}://{self.domain}{uri}')
        return Url.from_string(
            f'{self.scheme}://{self.domain}{self.path}/{uri}'
        )


class UrlId(Url):
    """Url id class.

    The url id class is used when creating a photo and
    it might not be neccessary to load all url properties
    yet, but might be neccessary in the future.
    """

    def __init__(self, id: str):
        """Create new url.

        Args:
            id: Url id (same as the hash of a real url)
        """
        self.id = id


class InvalidUrlException(ValueError):
    """Invalid url exception."""

    pass



"""Setup."""

from setuptools import setup, find_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
readme = path.join(this_directory, 'README.md')
with open(readme, encoding='utf-8') as file:
    long_description = file.read()

setup(
    name='saas',
    version='1.1.0.2',
    description='Screenshot as a service',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Ludwig Kristoffersson',
    author_email='ludwig@kristoffersson.org',
    url='https://github.com/nattvara/saas',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.7',
    install_requires=[
        'wheel==0.32.*',
        'beeprint==2.4.*',
        'elasticsearch==6.3.*',
        'selenium==3.141.*',
        'fusepy==3.0.*',
        'psutil==4.3.*',
    ],
    package_data={
        'extensions': [
            'idcac_2.9.5.xpi',
            'referer_header.xpi',
            'uBlock0_1.17.7rc0.xpi',
        ]
    },
    entry_points={
        'console_scripts': [
            'saas = saas.saas:main'
        ]
    }
)



"""Camera test."""

from saas.photographer.camera import Camera, UserAgents, Limits
from saas.photographer.javascript import JavascriptSnippets
from saas.storage.datadir import DataDirectory
from saas.photographer.photo import PhotoPath
from unittest.mock import MagicMock, call
from selenium import webdriver
from saas.web.url import Url
from os.path import dirname
import unittest
import time


class TestCamera(unittest.TestCase):
    """Test camera class."""

    def setUp(self):
        """Set up test."""
        self.camera = Camera()
        webdriver.Firefox.__init__ = MagicMock(return_value=None)
        self.url = Url.from_string('https://example.com')
        self.datadir = DataDirectory(dirname(__file__) + '/datadir')

    def tearDown(self):
        """Tear down test."""
        self.datadir.remove_data_dir()

    def creates_webdriver(self):
        """Test creates a webdriver."""
        self.webdriver_profile = self.camera._create_webdriver_profile()
        self.camera.webdriver = self.camera._create_webdriver(
            self.webdriver_profile,
            UserAgents.GOOGLEBOT
        )

    def routes_to_url(self):
        """Test routes camera to a url."""
        self.camera.webdriver.get = MagicMock()

    def uses_javascript_snippets(self):
        """Test uses javascript snippets."""
        JavascriptSnippets.load()

    def test_camera_can_create_webdriver_profile(self):
        """Test camerea can create webdriver profile."""
        self.creates_webdriver()
        self.assertIsInstance(
            obj=self.webdriver_profile,
            cls=webdriver.FirefoxProfile
        )

    def test_camera_can_create_webdriver(self):
        """Test camera can create webdriver."""
        self.creates_webdriver()
        self.webdriver_profile.set_preference = MagicMock()

        self.camera.webdriver = self.camera._create_webdriver(
            self.webdriver_profile,
            UserAgents.GOOGLEBOT
        )

        self.assertIsInstance(obj=self.camera.webdriver, cls=webdriver.Firefox)
        self.webdriver_profile.set_preference.assert_any_call(
            'general.useragent.override',
            UserAgents.GOOGLEBOT
        )
        self.webdriver_profile.set_preference.assert_any_call(
            'dom.popup_maximum',
            0
        )
        self.webdriver_profile.set_preference.assert_any_call(
            'privacy.popups.showBrowserMessage',
            False
        )

    def test_camera_can_route_to_url(self):
        """Test camera can be routed to url."""
        self.creates_webdriver()
        self.routes_to_url()

        self.camera._route(self.url)
        self.camera.webdriver.get.assert_called_with(self.url.to_string())

    def test_camera_resolution_can_be_set(self):
        """Test camera resolution can be set."""
        self.creates_webdriver()
        self.camera.webdriver.set_window_size = MagicMock()

        self.camera._set_resolution(100, 200)
        self.camera.webdriver.set_window_size.assert_called_with(100, 200)

    def test_camera_can_get_document_height(self):
        """Test camera can get document size."""
        self.creates_webdriver()
        self.uses_javascript_snippets()
        self.camera.webdriver.execute_script = MagicMock(return_value=10000)

        self.camera._document_height()
        self.camera.webdriver.execute_script.assert_called_with(
            JavascriptSnippets.DOCUMENT_HEIGHT
        )

    def test_camera_can_get_image_count(self):
        """Test camera can get image count."""
        self.creates_webdriver()
        self.uses_javascript_snippets()
        self.camera.webdriver.execute_script = MagicMock(return_value=123)

        images = self.camera._image_count()
        self.assertEqual(123, images)
        self.camera.webdriver.execute_script.assert_called_with(
            JavascriptSnippets.IMAGE_COUNT
        )

    def test_camera_can_get_script_count(self):
        """Test camera can get script count."""
        self.creates_webdriver()
        self.uses_javascript_snippets()
        self.camera.webdriver.execute_script = MagicMock(return_value=20)

        scripts = self.camera._script_count()
        self.assertEqual(20, scripts)
        self.camera.webdriver.execute_script.assert_called_with(
            JavascriptSnippets.SCRIPT_COUNT
        )

    def test_camera_can_get_stylesheet_count(self):
        """Test camera can get stylesheet count."""
        self.creates_webdriver()
        self.uses_javascript_snippets()
        self.camera.webdriver.execute_script = MagicMock(return_value=10)

        stylesheets = self.camera._stylesheet_count()
        self.assertEqual(10, stylesheets)
        self.camera.webdriver.execute_script.assert_called_with(
            JavascriptSnippets.STYLESHEET_COUNT
        )

    def test_camera_can_start_images_monitor(self):
        """Test camera can start images monitor."""
        self.creates_webdriver()
        self.uses_javascript_snippets()
        self.camera.webdriver.execute_script = MagicMock()

        self.camera._start_images_monitor()
        self.camera.webdriver.execute_script.assert_called_with(
            JavascriptSnippets.IMAGES_MONITOR
        )

    def test_camera_can_check_if_all_images_have_loaded(self):
        """Test camera can check if all images have finished loading."""
        self.creates_webdriver()
        self.uses_javascript_snippets()
        time.sleep = MagicMock()
        self.camera.webdriver.execute_script = MagicMock()
        self.camera.webdriver.execute_script.side_effect = [
            None,  # start images monitor
            0,
            0,
            0,
            1,
            1,
            2,
            3,
            4,
            Limits.ALL_IMAGES_LOADED,  # All images loaded at this limit
        ]

        self.camera._start_images_monitor()
        self.camera.webdriver.execute_script.assert_called_with(
            JavascriptSnippets.IMAGES_MONITOR
        )

        self.camera._wait_for_images_to_load()
        calls = [
            call(JavascriptSnippets.IMAGES_LOADED),
            call(JavascriptSnippets.IMAGES_LOADED),
            call(JavascriptSnippets.IMAGES_LOADED),
            call(JavascriptSnippets.IMAGES_LOADED),
            call(JavascriptSnippets.IMAGES_LOADED),
            call(JavascriptSnippets.IMAGES_LOADED),
            call(JavascriptSnippets.IMAGES_LOADED),
            call(JavascriptSnippets.IMAGES_LOADED),
            call(JavascriptSnippets.IMAGES_LOADED),
        ]
        self.camera.webdriver.execute_script.assert_has_calls(calls)

    def test_camera_can_scroll_page(self):
        """Test camera can scroll page."""
        self.creates_webdriver()
        self.uses_javascript_snippets()
        self.camera.webdriver.execute_script = MagicMock()

        self.camera._scroll_y_axis(100)  # scroll down 100px

        script = JavascriptSnippets.SCROLL_Y_AXIS.replace('Y_PIXELS', '100')
        self.camera.webdriver.execute_script.assert_called_with(
            script
        )

    def test_camera_can_wait_for_resize_of_document(self):
        """Test camera can wait for resize in proportion to doc height."""
        time.sleep = MagicMock()
        self.creates_webdriver()
        self.camera.webdriver.execute_script = MagicMock()
        self.camera.webdriver.set_window_size = MagicMock()
        self.camera._image_count = MagicMock(return_value=50)
        self.camera._script_count = MagicMock(return_value=3)
        self.camera._stylesheet_count = MagicMock(return_value=2)

        self.camera._set_resolution(200, 10000)
        self.camera._wait_for_resize()

        calls = [
            call(Limits.RESIZE_WAIT_TIME),
            call(Limits.RESIZE_WAIT_TIME),
            call(Limits.RESIZE_WAIT_TIME),
            call(Limits.RESIZE_WAIT_TIME),
            call(Limits.RESIZE_WAIT_TIME),
            call(Limits.RESIZE_WAIT_TIME),
        ]
        time.sleep.assert_has_calls(calls)

    def test_camera_can_save_screenshot(self):
        """Test camera can save screenshot."""
        self.creates_webdriver()
        self.camera.webdriver.save_screenshot = MagicMock()

        path = PhotoPath(self.datadir)
        self.camera._save(path)

        self.camera.webdriver.save_screenshot.assert_called_with(
            path.full_path()
        )


if __name__ == '__main__':
    unittest.main()



"""File module."""

from __future__ import annotations
from saas.storage.refresh import RefreshRate
import saas.storage.index as idx
from typing import Optional, Type
import time
import os


class Path:
    """Path class."""

    RENDERING_EXTENSION = '.rendering.saas'

    def __init__(self, path: str):
        """Create path.

        Args:
            path: a string representing a path, for example
                /example.com/2019-01-13H20:00/some_dir/index.png
        """
        if len(path) == 0:
            raise InvalidPathException(
                f'path was invalid, must be longer than 0 chars'
            )

        if path[0] != '/':
            raise InvalidPathException('path must start with a slash \'/\'')

        path = path.replace(Path.RENDERING_EXTENSION, '')

        self.domain = self._parse_domain(path)
        self.captured_at = self._parse_captured_at(path)
        self.end = self._parse_end(path)

    def _parse_domain(self, path: str) -> str:
        """Parse domain from path.

        Args:
            path: Path to parse

        Returns:
            A domain name
            str
        """
        pieces = path.split('/')
        return pieces[1]

    def _parse_captured_at(self, path: str) -> str:
        """Parse captured_at from path.

        Args:
            path: A path to parse

        Returns:
            A captured_at value (see refresh module)
            str
        """
        pieces = path.split('/')
        if len(pieces) >= 3:
            return pieces[2]
        return ''

    def _parse_end(self, path: str) -> str:
        """Parse end from a path.

        Args:
            path: A path to parse

        Returns:
            Everything from path except from domain and
                captured_at
            str
        """
        pieces = path.split('/')
        piece = pieces[3:]
        path = '/' + '/'.join(piece)
        if len(piece) > 1:
            path = path.rstrip('/')
        return path

    def includes_domain(self) -> bool:
        """Check if path includes domain.

        Returns:
            True if path includes domain, otherwise False
            bool
        """
        if self.domain != '':
            return True
        return False

    def includes_captured_at(self) -> bool:
        """Check if path includes captured_at.

        Returns:
            True if path includes captured at, otherwise False
            bool
        """
        if self.captured_at != '':
            return True
        return False

    def includes_end(self) -> bool:
        """Check if path includes end.

        Returns:
            True if path includes end, otherwise False
            bool
        """
        if self.end == '':
            return False
        if self.end == '/':
            return False
        return True

    def end_as_file(self) -> str:
        """Get end of path as a file.

        Should never end with a slash

        Returns:
            End of path
            str
        """
        return self.end

    def end_as_directory(self) -> str:
        """Get end of path as a directory.

        Should always end with a slash

        Returns:
            End of path
            str
        """
        end = self.end.rstrip('/')
        end += '/'
        return end


class InvalidPathException(Exception):
    """Invalid path exception."""

    pass


class Directory:
    """Directory class."""

    ST_MODE = 0o40755

    TIME = 0.0

    def __init__(self, filename: str):
        """Create new directory.

        Args:
            filename: directory filename
        """
        self.filename = filename
        self.st_mode = Directory.ST_MODE  # Permissions
        if Directory.TIME == 0.0:
            Directory.TIME = time.time()

    def attributes(self: Optional['Directory']=None) -> dict:
        """Get attributes of file.

        Args:
            self: Self (default: {None})

        Returns:
            File attributes
            dict
        """
        return {
            'st_atime': Directory.TIME,
            'st_ctime': Directory.TIME,
            'st_gid': os.getgid(),
            'st_mode': Directory.ST_MODE,
            'st_mtime': Directory.TIME,
            'st_size': 0,
            'st_uid': os.getuid(),
        }


class File:
    """File class."""

    ST_MODE = 0o100644

    TIME = 0.0

    def __init__(self, filename: str):
        """Create new file.

        Args:
            filename: file's filename
        """
        self.filename = filename
        self.st_mode = File.ST_MODE  # Permissions
        if File.TIME == 0.0:
            File.TIME = time.time()

    def attributes(self: Optional['File']=None, filesize: int=0) -> dict:
        """Get attributes of file.

        Args:
            self: Self (default: {None})
            filesize: size of file on disk (default: {0})

        Returns:
            File attributes
            dict
        """
        return {
            'st_atime': File.TIME,
            'st_ctime': File.TIME,
            'st_gid': os.getgid(),
            'st_mode': File.ST_MODE,
            'st_mtime': File.TIME,
            'st_size': filesize,
            'st_uid': os.getuid(),
        }


class LastCapture:
    """Last captured class.

    The last capture is the most recent capture of a domain.
    This class helps convert last capture placeholder to
    real captured_at value.
    """

    FILENAME = 'latest'

    CACHE_AGE_LIMIT = 60

    captures = {}  # type: dict

    cached_at = {}  # type: dict

    @staticmethod
    def translate(
        captured_at: str,
        domain: str,
        index: idx.Index,
        refresh_rate: Type[RefreshRate]
    ) -> str:
        """Translate captured_at to cached value.

        If a captured_at value is FILENAME this will translate
        it to the most recent captured_at value for given
        domain and refresh_rate

        Args:
            captured_at: photos captured at
            domain: photos from a given domain
            index: Index photos are stored in
            refresh_rate: refresh rate capture should be for

        Returns:
            original value if captured_at was not FILENAME, if it
            was FILENAME then the most recent captured_at value is returned
            str
        """
        try:
            if captured_at == LastCapture.FILENAME:
                if not LastCapture._is_cached(domain):
                    LastCapture._update(domain, index, refresh_rate)
                captured_at = LastCapture._from_cache(domain)
        except idx.EmptySearchResultException:
            pass
        return captured_at

    @staticmethod
    def _from_cache(domain: str) -> str:
        """Get last capture of domain from cache.

        Args:
            domain: last capture of domain

        Returns:
            Last captured_at value
            str
        """
        captured_at = LastCapture.captures[domain]  # type: str
        return captured_at

    @staticmethod
    def _is_cached(domain: str) -> bool:
        """Check if last capture has a valid cache.

        Args:
            domain: last capture of domain

        Returns:
            True if a valid cache exists, otherwise False
            bool
        """
        if domain not in LastCapture.captures:
            return False

        cache_age = time.time() - LastCapture.cached_at[domain]

        if cache_age > LastCapture.CACHE_AGE_LIMIT:
            return False

        return True

    @staticmethod
    def _update(
        domain: str,
        index: idx.Index,
        refresh_rate: Type[RefreshRate]
    ):
        """Update cache.

        Args:
            domain: domain to cache
            index: Index photos are stored in
            refresh_rate: refresh rate capture should be for
        """
        capture = index.photos_most_recent_capture_of_domain(
            domain,
            refresh_rate
        )
        LastCapture.captures[domain] = capture
        LastCapture.cached_at[domain] = time.time()



"""Photographer test."""

from saas.photographer.photographer import Photographer
from saas.storage.datadir import DataDirectory
import saas.storage.refresh as refresh
from saas.storage.index import Index
from unittest.mock import MagicMock
from saas.web.url import Url
from os.path import dirname
import unittest


class TestPhotographer(unittest.TestCase):
    """Test photographer class."""

    def setUp(self):
        """Set up test."""
        self.index = Index()
        self.datadir = DataDirectory(dirname(__file__) + '/datadir')
        self.photographer = Photographer(
            self.index,
            refresh.Hourly,
            self.datadir
        )

    def tearDown(self):
        """Tear down test."""
        self.datadir.remove_data_dir()

    def does_url_checkout(self):
        """Test does checkut of url.

        Add mock to affected index methods.
        """
        url = Url.from_string('https://example.com')
        self.index.recently_crawled_url = MagicMock(return_value=url)
        self.index.crawled_urls_count = MagicMock(return_value=1)
        self.index.lock_crawled_url = MagicMock()

    def test_photographer_can_checkout_url_from_crawled_index(self):
        """Test photographer can checkout url from "crawled" index."""
        self.does_url_checkout()

        url = self.photographer._checkout_url()
        self.assertIsInstance(cls=Url, obj=url)

    def test_photographer_locks_the_url_on_checkout(self):
        """Test photographer locks url on checkout."""
        self.does_url_checkout()

        url = self.photographer._checkout_url()
        self.index.lock_crawled_url.assert_called_with(
            url,
            refresh.Hourly
        )


if __name__ == '__main__':
    unittest.main()



"""Photo module."""

from __future__ import annotations
import saas.storage.datadir as DataDirectory
import saas.storage.refresh as refresh
from saas.web.url import Url
from abc import ABCMeta
from typing import Type
import uuid
import os


class Photo(metaclass=ABCMeta):
    """Base Photo class."""

    def __init__(
        self,
        url: Url,
        path: 'PhotoPath',
        refresh_rate: Type[refresh.RefreshRate],
        index_filesize: int=None
    ):
        """Create new photo.

        Args:
            url: The photo is taken of given Url
            path: Path to photo in data directory
            refresh_rate: The refresh rate of the photo (hourly, daily, etc.)
            index_filesize: If photo have been stored in index, filesize is
                already stored there. To speed up performance this takes
                priority over filesize in datadir. see self.filesize()
        """
        self.url = url
        self.path = path
        self.refresh_rate = refresh_rate
        self.index_filesize = index_filesize

    def get_raw(self) -> str:
        """Get raw content of photos file in data directory.

        Returns:
            Raw source
            str
        """
        file = open(self.path.full_path(), 'r')
        content = file.read()
        file.close()
        return content

    def filename(self) -> str:
        """Get photo filename.

        Returns:
            A filename based on the photos url
            str
        """
        return self.url.make_filename()

    def directory(self) -> str:
        """Get photo directory.

        Returns:
            A directory based on the photos url
            str
        """
        return self.url.make_directory()

    def domain(self) -> str:
        """Get photo domain.

        Returns:
            Domain photo belongs to
            str
        """
        return self.url.domain

    def filesize(self) -> int:
        """Get photo filesize.

        Returns:
            Size in bytes
            int
        """
        if self.index_filesize:
            return self.index_filesize
        return self.path.filesize()


class LoadingPhoto(Photo):
    """Loading photo class.

    A loading photo is created once photographer do a
    checkout of a url. This is used to display a webpage
    that is being rendered in the mounted filesystem.
    """

    def save_loading_text(self):
        """Write loading text to photo source."""
        file = open(self.path.full_path(), 'w+')
        file.write('loading')
        file.close()


class Screenshot(Photo):
    """Screenshot photo class."""

    pass


class PhotoPath:
    """Photopath class."""

    def __init__(
        self,
        datadir: DataDirectory.DataDirectory,
        uuid: str=''
    ):
        """Create new path to a photo.

        Args:
            datadir: Data directory to store photo in
            uuid: If an PhotoPath should represent an existing
                path, use uuid argument, otherwise a new uuid
                will be generated
        """
        self.datadir = datadir
        if uuid != '':
            self.uuid = uuid
        else:
            self.uuid = self.make_uuid()

    def make_uuid(self) -> str:
        """Make uuid.

        Returns:
            A unique id
            str
        """
        return str(uuid.uuid4())

    def full_path(self) -> str:
        """Get full path to photo's file in datadir.

        Returns:
            An absolute path to the photo
            str
        """
        return self.datadir.path_for_photo(self)

    def filesize(self) -> int:
        """Get filesize.

        Returns:
            Size in bytes
            int
        """
        return os.path.getsize(self.full_path())

    def should_optimize(self) -> bool:
        """If photo file should be optimized.

        Returns:
            True if should be optimize, otherwise False
            bool
        """
        return self.datadir.optimize_storage

    def optimize(self):
        """Optimize png file in data directory."""
        self.datadir.optimize_file(self.full_path())
